#!/usr/bin/python

# A simple three-species chemical kinetics system known as
# "Robertson's example", as presented in:
#
# H. H. Robertson, The solution of a set of reaction rate equations,
# in Numerical Analysis: An Introduction, J. Walsh, ed., Academic
# Press, 1966, pp. 178-182.
#
# This is a simple system often used to study stiffness in systems of
# differential equations.  It doesn't leverage the power of
# rules-based modeling or pysb, but it's a useful small model for
# purposes of experimentation.
#
# A brief report addressing issues of stiffness encountered in
# numerical integration of Robertson's example can be found here:
# http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.53.8603&rep=rep1&type=pdf
#
# The chemical model is as follows:
#
#      Reaction        Rate
#   ------------------------
#       A -> B         0.04
#      2B -> B + C     3.0e7
#   B + C -> A + C     1.0e4
#
# The resultant system of differential equations is:
#
# y1' = -0.04 * y1 + 1.0e4 * y2 * y3
# y2' =  0.04 * y1 - 1.0e4 * y2 * y3 - 3.0e7 * y2^2
# y3' =                                3.0e7 * y2^2
#
# If you run this script directly, it will generate the equations and
# print them in a form very closely resembling that given above.  See
# also robertson_simulate.py which integrates the system and plots the
# trajectories.


from pysb import *

Model()

# rates
Parameter('k1', 0.04)
Parameter('k2', 3.0e7)
Parameter('k3', 1.0e4)
# initial conditions
Parameter('A_0', 1.0)
Parameter('B_0', 0.0)
Parameter('C_0', 0.0)

Monomer('A')
Monomer('B')
Monomer('C')

Rule('A_to_B', A() >> B(), k1)
Rule('BB_to_BC', B() + B() >> B() + C(), k2)
Rule('BC_to_AC', B() + C() >> A() + C(), k3)

Initial(A(), A_0)
Initial(B(), B_0)
Initial(C(), C_0)


if __name__ == '__main__':
    from pysb.bng import generate_equations
    # this creates model.odes which contains the math
    generate_equations(model)

    # here we build a symbol substitution mapping to help the math
    # look like the equations in the comment above, instead of showing
    # the internal pysb symbol names
    substitutions = {}
    # map parameter symbols to their real values
    substitutions.update((name, param.value) for name, param in model.parameters.items())
    # map species variables sI to yI+1, e.g. s0 -> y1
    substitutions.update(('s%d' % i, 'y%d' % (i+1)) for i in range(len(model.odes)))

    print "Generated differential equations for Roberson's example:"
    print
    # iterate over each equation
    for i, eq in enumerate(model.odes):
        # perform the substitution usign the mappings we built above
        eq_sub = eq.subs(substitutions)
        # display the equation
        print 'y%d\' = %s' % (i+1, eq_sub)



from pysb.integrate import odesolve
from pylab import *

from earm_1_0 import model


# saturating level of ligand (corresponding to ~1000 ng/ml SuperKiller TRAIL)
Lsat = 6E4; 

# relationship of ligand concentration in the model (in # molecules/cell) to actual TRAIL concentration (in ng/ml)
Lfactor = model.parameters['L_0'].value / 50;

L_0_baseline = model.parameters['L_0'].value


def fig_4a():
    t = linspace(0, 20*3600, 20*60+1)  # 20 hours, in seconds, 1 min sampling
    dt = t[1] - t[0]

    Ls_exp = Lsat / array([1, 4, 20, 100, 500])
    Td_exp = [144.2, 178.7, 236,   362.5, 656.5]
    Td_std = [32.5,   32.2,  36.4,  78.6, 171.6]
    Ts_exp = [21.6,   23.8,  27.2,  22.0,  19.0]
    Ts_std = [9.5,     9.5,  12.9,   7.7,  10.5]

    CVenv = 0.2
    # num steps was originally 40, but 15 is plenty smooth enough for screen display
    Ls = floor(logspace(1,5,15)) 

    fs = empty_like(Ls)
    Ts = empty_like(Ls)
    Td = empty_like(Ls)
    for i in range(len(Ls)):
        model.parameters['L_0'].value = Ls[i]

        print "integrating model"
        x = odesolve(model, t)

        fs[i] = (x['PARP'][0] - x['PARP'][-1]) / x['PARP'][0]
        dP = 60 * (x['PARP'][:-1] - x['PARP'][1:]) / (dt * x['PARP'][0])  # in minutes
        ttn = argmax(dP)
        dPmax = dP[ttn]
        Ts[i] = 1 / dPmax  # minutes
        Td[i] = t[ttn] / 60  # minutes

    figure()
    plot(Ls/Lfactor, Td, 'g-', Ls/Lfactor, (1-CVenv)*Td, 'g--', Ls/Lfactor, (1+CVenv)*Td, 'g--')
    errorbar(Ls_exp/Lfactor, Td_exp, Td_std, None, 'go', capsize=0),
    ylabel('Td (min)'),
    xlabel('TRAIL (ng/ml)'),
    a = gca()
    a.set_xscale('log')
    a.set_xlim((min(Ls) / Lfactor, max(Ls) / Lfactor))
    a.set_ylim((0, 1000))

    model.parameters['L_0'].value = L_0_baseline


def fig_4b():
    t = linspace(0, 6*3600, 6*60+1)  # 6 hours
    x = odesolve(model, t)

    x_norm = array([x['Bid'], x['PARP'], x['mSmac']]).T
    x_norm = 1 - x_norm / x_norm[0, :]  # gets away without max() since first values are largest

    # this is what I originally thought 4B was plotting.  oddly it's very close.
    #x_norm = array([x['tBid'], x['CPARP'], x['cSmac']]).T
    #x_norm /= x_norm.max(0)

    tp = t / 3600  # x axis as hours

    figure()
    plot(tp, x_norm[:,0], 'b', label='IC substrate (tBid)')
    plot(tp, x_norm[:,1], 'y', label='EC substrate (cPARP)')
    plot(tp, x_norm[:,2], 'r', label='MOMP (cytosolic Smac)')
    legend(loc='upper left', bbox_to_anchor=(0,1)).draw_frame(False)
    xlabel('Time (hr)')
    ylabel('fraction')
    a = gca()
    a.set_ylim((-.05, 1.05))


fig_4a()
fig_4b()
show()



from pylab import *
from pysb.integrate import odesolve

from bax_tetramer import model


t = linspace(0, 5e4)
x = odesolve(model, t)

p = plot(t, array([x['BAX4'], x['BAX4_inh']]).T)
figlegend(p, ['BAX4', 'BAX4_inh'], 'upper left')
show()



import sys
import os
import errno
import warnings
import inspect
import re
import collections

def Observe(*args):
    return SelfExporter.default_model.add_observable(*args)

def Initial(*args):
    return SelfExporter.default_model.initial(*args)

def MatchOnce(pattern):
    cp = as_complex_pattern(pattern).copy()
    cp.match_once = True
    return cp


# Internal helper to implement the magic of making model components
# appear in the calling module's namespace.  Do not construct any
# instances; we just use the class for namespace containment.
class SelfExporter(object):

    do_export = True
    default_model = None
    target_globals = None   # the globals dict to which we'll export our symbols
    target_module = None    # the module to which we've exported

    @staticmethod
    def export(obj):
        if not SelfExporter.do_export:
            return
        if not isinstance(obj, (Model, Component)):
            raise Exception("%s is not a type that is understood by SelfExporter" % str(type(obj)))

        # determine the module from which we were called (we need to do this here so we can
        # calculate stacklevel for use in the warning at the bottom of this method)
        cur_module = inspect.getmodule(inspect.currentframe())
        caller_frame = inspect.currentframe()
        # walk up through the stack until we hit a different module
        stacklevel = 1
        while inspect.getmodule(caller_frame) == cur_module:
            stacklevel += 1
            caller_frame = caller_frame.f_back

        # use obj's name as the symbol to export it to (unless modified below)
        export_name = obj.name

        if isinstance(obj, Model):
            if SelfExporter.default_model is not None:
                warnings.warn("Redefining model! (You can probably ignore this if you are running"
                              " code interactively)", ModelExistsWarning, stacklevel);
                # delete previously exported symbols to prevent extra SymbolExistsWarnings
                for name in [c.name for c in SelfExporter.default_model.all_components()] + ['model']:
                    if name in SelfExporter.target_globals:
                        del SelfExporter.target_globals[name]
            SelfExporter.target_module = inspect.getmodule(caller_frame)
            SelfExporter.target_globals = caller_frame.f_globals
            SelfExporter.default_model = obj
            # if not set, assign model's name from the module it lives in. very sneaky and fragile.
            if obj.name is None:
                if SelfExporter.target_module == sys.modules['__main__']:
                    # user ran model .py directly
                    model_filename = inspect.getfile(sys.modules['__main__'])
                    module_name = re.sub(r'\.py$', '', model_filename)
                elif SelfExporter.target_module is not None:
                    # model is imported by some other script (typical case)
                    module_name = SelfExporter.target_module.__name__
                else:
                    # user is defining a model interactively (not really supported, but we'll try)
                    module_name = '<interactive>'
                obj.name = module_name   # internal name for identification
                export_name = 'model'    # symbol name for export
        elif isinstance(obj, Component):
            if SelfExporter.default_model == None:
                raise Exception("A Model must be declared before declaring any model components")
            SelfExporter.default_model.add_component(obj)

        # load obj into target namespace under obj.name
        if SelfExporter.target_globals.has_key(export_name):
            warnings.warn("'%s' already defined" % (export_name), SymbolExistsWarning, stacklevel)
        SelfExporter.target_globals[export_name] = obj


class Component(object):
    """The base class for all the things contained within a model."""

    def __init__(self, name, _export=True):
        if not re.match(r'[_a-z][_a-z0-9]*\Z', name, re.IGNORECASE):
            raise InvalidComponentNameError(name)
        self.name = name
        if _export:
            try:
                SelfExporter.export(self)
            except ComponentDuplicateNameError as e:
                # re-raise to hide the stack trace below this point -- it's irrelevant to the user
                # and makes the error harder to understand
                raise e


class Model(object):

    def __init__(self, name=None, _export=True):
        self.name = name
        self.monomers = ComponentSet()
        self.compartments = ComponentSet()
        self.parameters = ComponentSet()
        self.rules = ComponentSet()
        self.species = []
        self.odes = []
        self.reactions = []
        self.reactions_bidirectional = []
        self.observable_patterns = []
        self.observable_groups = {}  # values are tuples of factor,speciesnumber
        self.initial_conditions = []
        if _export:
            SelfExporter.export(self)

    def reload(self):
        # forcibly removes the .pyc file and reloads the model module
        model_pyc = SelfExporter.target_module.__file__
        if model_pyc[-3:] == '.py':
            model_pyc += 'c'
        try:
            os.unlink(model_pyc)
        except OSError as e:
            # ignore "no such file" errors, re-raise the rest
            if e.errno != errno.ENOENT:
                raise
        try:
            reload(SelfExporter.target_module)
        except SystemError as e:
            # This one specific SystemError occurs when using ipython to 'run' a model .py file
            # directly, then reload()ing the model, which makes no sense anyway. (just re-run it)
            if e.args == ('nameless module',):
                raise Exception('Cannot reload a model which was executed directly in an interactive'
                                'session. Please import the model file as a module instead.')
            else:
                raise
        # return self for "model = model.reload()" idiom, until a better solution can be found
        return SelfExporter.default_model

    def all_components(self):
        components = ComponentSet()
        for container in [self.monomers, self.compartments, self.rules, self.parameters]:
            components |= container
        return components

    def parameters_rules(self):
        """Returns a ComponentSet of the parameters used as rate constants in rules"""
        # rate_reverse is None for irreversible rules, so we'll need to filter those out
        cset = ComponentSet(p for r in self.rules for p in (r.rate_forward, r.rate_reverse)
                            if p is not None)
        # intersect with original parameter list to retain ordering
        return self.parameters & cset

    def parameters_initial_conditions(self):
        """Returns a ComponentSet of the parameters used as initial conditions"""
        cset = ComponentSet(ic[1] for ic in self.initial_conditions)
        # intersect with original parameter list to retain ordering
        return self.parameters & cset

    def parameters_compartments(self):
        """Returns a ComponentSet of the parameters used as compartment sizes"""
        cset = ComponentSet(c.size for c in self.compartments)
        # intersect with original parameter list to retain ordering
        return self.parameters & cset

    def parameters_unused(self):
        """Returns a ComponentSet of the parameters not used in the model at all"""
        cset_used = self.parameters_rules() | self.parameters_initial_conditions() | self.parameters_compartments()
        return self.parameters - cset_used

    def add_component(self, other):
        # We have 4 containers for the 4 types of components. This code determines the right one
        # based on the class of the object being added.  It tries to be defensive against reasonable
        # errors, but still seems sort of fragile.
        container_name = type(other).__name__.lower() + 's'
        container = getattr(self, container_name, None)
        if not isinstance(other, Component) or not isinstance(container, ComponentSet):
            raise Exception("Tried to add component of unknown type '%s' to model" % type(other))
        container.add(other)

    def add_observable(self, name, reaction_pattern):
        try:
            reaction_pattern = as_reaction_pattern(reaction_pattern)
        except InvalidReactionPatternException as e:
            raise type(e)("Observable pattern does not look like a ReactionPattern")
        self.observable_patterns.append( (name, reaction_pattern) )

    def initial(self, complex_pattern, value):
        try:
            complex_pattern = as_complex_pattern(complex_pattern)
        except InvalidComplexPatternException as e:
            raise type(e)("Initial condition species does not look like a ComplexPattern")
        if not isinstance(value, Parameter):
            raise Exception("Value must be a Parameter")
        if not complex_pattern.is_concrete():
            raise Exception("Pattern must be concrete")
        if any(complex_pattern.is_equivalent_to(other_cp) for other_cp, value in self.initial_conditions):
            # FIXME until we get proper canonicalization this could produce false negatives
            raise Exception("Duplicate initial condition")
        self.initial_conditions.append( (complex_pattern, value) )

    def get_species_index(self, complex_pattern):
        # FIXME I don't even want to think about the inefficiency of this, but at least it works
        try:
            return (i for i, s_cp in enumerate(self.species) if s_cp.is_equivalent_to(complex_pattern)).next()
        except StopIteration:
            return None

    def __repr__(self): 
        return "<%s '%s' (monomers: %d, rules: %d, parameters: %d, compartments: %d) at 0x%x>" % \
            (self.__class__.__name__, self.name, len(self.monomers), len(self.rules),
             len(self.parameters), len(self.compartments), id(self))



class Monomer(Component):
    """The Monomer class creates monomers with the specified sites, state-sites, and compartment
    """

    def __init__(self, name, sites=[], site_states={}, _export=True):
        Component.__init__(self, name, _export)

        # ensure sites is some kind of list (presumably of strings) but not a string itself
        if not isinstance(sites, collections.Iterable) or isinstance(sites, basestring):
            raise ValueError("sites must be a list of strings")
        
        # ensure no duplicate sites
        sites_seen = {}
        for site in sites:
            sites_seen.setdefault(site, 0)
            sites_seen[site] += 1
        sites_dup = [site for site in sites_seen.keys() if sites_seen[site] > 1]
        if sites_dup:
            raise Exception("Duplicate sites specified: " + str(sites_dup))

        # ensure site_states keys are all known sites
        unknown_sites = [site for site in site_states.keys() if not site in sites_seen]
        if unknown_sites:
            raise Exception("Unknown sites in site_states: " + str(unknown_sites))
        # ensure site_states values are all strings
        invalid_sites = [site for (site, states) in site_states.items() if not all([type(s) == str for s in states])]
        if invalid_sites:
            raise Exception("Non-string state values in site_states for sites: " + str(invalid_sites))

        self.sites = list(sites)
        self.sites_dict = dict.fromkeys(sites)
        self.site_states = site_states

    def __call__(self, *args, **kwargs):
        """Build a MonomerPattern object with convenient kwargs for the sites"""
        return MonomerPattern(self, extract_site_conditions(*args, **kwargs), None)

    def __repr__(self):
        return  '%s(name=%s, sites=%s, site_states=%s)' % \
            (self.__class__.__name__, repr(self.name), repr(self.sites), repr(self.site_states))

    

class MonomerAny(Monomer):

    def __init__(self):
        # don't call Monomer.__init__ since this doesn't want
        # Component stuff and has no user-accessible API
        self.name = 'ANY'
        self.sites = None
        self.sites_dict = {}
        self.site_states = {}
        self.compartment = None

    def __repr__(self):
        return self.name



class MonomerWild(Monomer):

    def __init__(self):
        # don't call Monomer.__init__ since this doesn't want
        # Component stuff and has no user-accessible API
        self.name = 'WILD'
        self.sites = None
        self.sites_dict = {}
        self.site_states = {}
        self.compartment = None

    def __repr__(self):
        return self.name



class MonomerPattern(object):

    def __init__(self, monomer, site_conditions, compartment):
        # ensure all keys in site_conditions are sites in monomer
        unknown_sites = [site for site in site_conditions.keys() if site not in monomer.sites_dict]
        if unknown_sites:
            raise Exception("MonomerPattern with unknown sites in " + str(monomer) + ": " + str(unknown_sites))

        # ensure each value is one of: None, integer, list of integers, string, (string,integer), (string,WILD), ANY
        # FIXME: support state sites
        invalid_sites = []
        for (site, state) in site_conditions.items():
            # pass through to next iteration if state type is ok
            if state == None:
                continue
            elif type(state) == int:
                continue
            elif type(state) == list and all(isinstance(s, int) for s in state):
                continue
            elif type(state) == str:
                continue
            elif type(state) == tuple and type(state[0]) == str and (type(state[1]) == int or state[1] == WILD):
                continue
            elif state == ANY:
                continue
            invalid_sites.append(site)
        if invalid_sites:
            raise Exception("Invalid state value for sites: " + '; '.join(['%s=%s' % (s,str(site_conditions[s])) for s in invalid_sites]))

        # ensure compartment is a Compartment
        if compartment and not isinstance(compartment, Compartment):
            raise Exception("compartment is not a Compartment object")

        self.monomer = monomer
        self.site_conditions = site_conditions
        self.compartment = compartment

    def is_concrete(self):
        """Return a bool indicating whether the pattern is 'concrete'.

        'Concrete' means the pattern satisfies ALL of the following:
        1. All sites have specified conditions
        2. If the model uses compartments, the compartment is specified.

        """
        # 1.
        sites_ok = self.is_site_concrete()
        # 2.
        # FIXME accessing the model via SelfExporter.default_model is a temporary hack - all model
        #   components (Component subclasses?) need weak refs to their parent model.
        compartment_ok = self.compartment is not None or not SelfExporter.default_model.compartments
        return compartment_ok and sites_ok

    def is_site_concrete(self):
        """Return a bool indicating whether the pattern is 'site-concrete'.

        'Site-concrete' means all sites have specified conditions."""
        # assume __init__ did a thorough enough job of error checking that this is is all we need to do
        return len(self.site_conditions) == len(self.monomer.sites)

    def __call__(self, *args, **kwargs):
        """Build a new MonomerPattern with updated site conditions. Can be used
        to obtain a shallow copy by passing an empty argument list."""
        # The new object will have references to the original monomer and
        # compartment, and a shallow copy of site_conditions which has been
        # updated according to our args (as in Monomer.__call__).
        site_conditions = self.site_conditions.copy()
        site_conditions.update(extract_site_conditions(*args, **kwargs))
        return MonomerPattern(self.monomer, site_conditions, self.compartment)

    def __add__(self, other):
        if isinstance(other, MonomerPattern):
            return ReactionPattern([ComplexPattern([self], None), ComplexPattern([other], None)])
        if isinstance(other, ComplexPattern):
            return ReactionPattern([ComplexPattern([self], None), other])
        else:
            return NotImplemented

    def __mod__(self, other):
        if isinstance(other, MonomerPattern):
            return ComplexPattern([self, other], None)
        else:
            return NotImplemented

    def __rshift__(self, other):
        if isinstance(other, (MonomerPattern, ComplexPattern, ReactionPattern)):
            return (self, other, False)
        else:
            return NotImplemented

    def __ne__(self, other):
        if isinstance(other, (MonomerPattern, ComplexPattern, ReactionPattern)):
            return (self, other, True)
        else:
            return NotImplemented

    def __pow__(self, other):
        if isinstance(other, Compartment):
            mp_new = self()
            mp_new.compartment = other
            return mp_new
        else:
            return NotImplemented

    def __repr__(self):
        value = '%s(' % self.monomer.name
        value += ', '.join([
                k + '=' + str(self.site_conditions[k])
                for k in self.monomer.sites
                if self.site_conditions.has_key(k)
                ])
        value += ')'
        if self.compartment is not None:
            value += ' ** ' + self.compartment.name
        return value



class ComplexPattern(object):
    """Represents a bound set of MonomerPatterns, i.e. a complex.  In
    BNG terms, a list of patterns combined with the '.' operator)."""

    def __init__(self, monomer_patterns, compartment, match_once=False):
        # ensure compartment is a Compartment
        if compartment and not isinstance(compartment, Compartment):
            raise Exception("compartment is not a Compartment object")

        self.monomer_patterns = monomer_patterns
        self.compartment = compartment
        self.match_once = match_once

    def is_concrete(self):
        """Return a bool indicating whether the pattern is 'concrete'.

        'Concrete' means the pattern satisfies ANY of the following:
        1. All monomer patterns are concrete
        2. The compartment is specified AND all monomer patterns are site-concrete

        """
        # 1.
        mp_concrete_ok = all(mp.is_concrete() for mp in self.monomer_patterns)
        # 2.
        compartment_ok = self.compartment is not None and \
            all(mp.is_site_concrete() for mp in self.monomer_patterns)
        return mp_concrete_ok or compartment_ok

    def is_equivalent_to(self, other):
        """Checks for equality with another ComplexPattern"""
        # Didn't implement __eq__ to avoid confusion with __ne__ operator used for Rule building

        # FIXME the literal site_conditions comparison requires bond numbering to be identical,
        #   so some sort of canonicalization of that numbering is necessary.
        if not isinstance(other, ComplexPattern):
            raise Exception("Can only compare ComplexPattern to another ComplexPattern")
        return \
            sorted((mp.monomer, mp.site_conditions) for mp in self.monomer_patterns) == \
            sorted((mp.monomer, mp.site_conditions) for mp in other.monomer_patterns)

    def copy(self):
        """Implement our own brand of shallow copy.

        The new object will have references to the original compartment, and
        copies of the monomer_patterns."""
        return ComplexPattern([mp() for mp in self.monomer_patterns], self.compartment, self.match_once)

    def __add__(self, other):
        if isinstance(other, ComplexPattern):
            return ReactionPattern([self, other])
        elif isinstance(other, MonomerPattern):
            return ReactionPattern([self, ComplexPattern([other], None)])
        else:
            return NotImplemented

    def __mod__(self, other):
        if isinstance(other, MonomerPattern):
            return ComplexPattern(self.monomer_patterns + [other], self.compartment, self.match_once)
        elif isinstance(other, ComplexPattern):
            if self.compartment is not other.compartment:
                raise ValueError("merged ComplexPatterns must specify the same compartment")
            elif self.match_once != other.match_once:
                raise ValueError("merged ComplexPatterns must have the same value of match_once")
            return ComplexPattern(self.monomer_patterns + other.monomer_patterns, self.compartment, self.match_once)
        else:
            return NotImplemented

    def __rshift__(self, other):
        if isinstance(other, (MonomerPattern, ComplexPattern, ReactionPattern)):
            return (self, other, False)
        else:
            return NotImplemented

    def __ne__(self, other):
        if isinstance(other, (MonomerPattern, ComplexPattern, ReactionPattern)):
            return (self, other, True)
        else:
            return NotImplemented

    def __pow__(self, other):
        if isinstance(other, Compartment):
            cp_new = self.copy()
            cp_new.compartment = other
            return cp_new
        else:
            return NotImplemented

    def __repr__(self):
        ret = ' % '.join([repr(p) for p in self.monomer_patterns])
        if self.compartment is not None:
            ret = '(%s) ** %s' % (ret, self.compartment.name)
        if self.match_once:
            ret = 'MatchOnce(%s)' % ret
        return ret



class ReactionPattern(object):
    """Represents a complete pattern for the product or reactant side
    of a rule.  Essentially a thin wrapper around a list of
    ComplexPatterns."""

    def __init__(self, complex_patterns):
        self.complex_patterns = complex_patterns

    def __add__(self, other):
        if isinstance(other, MonomerPattern):
            return ReactionPattern(self.complex_patterns + [ComplexPattern([other], None)])
        elif isinstance(other, ComplexPattern):
            return ReactionPattern(self.complex_patterns + [other])
        else:
            return NotImplemented

    def __rshift__(self, other):
        """Irreversible reaction"""
        if isinstance(other, (MonomerPattern, ComplexPattern, ReactionPattern)):
            return (self, other, False)
        else:
            return NotImplemented

    def __ne__(self, other):
        """Reversible reaction"""
        if isinstance(other, (MonomerPattern, ComplexPattern, ReactionPattern)):
            return (self, other, True)
        else:
            return NotImplemented

    def __repr__(self):
        return ' + '.join([repr(p) for p in self.complex_patterns])



def as_complex_pattern(v):
    """Internal helper to 'upgrade' a MonomerPattern to a ComplexPattern."""

    if isinstance(v, ComplexPattern):
        return v
    elif isinstance(v, MonomerPattern):
        return ComplexPattern([v], None)
    else:
        raise InvalidComplexPatternException


def as_reaction_pattern(v):
    """Internal helper to 'upgrade' a Complex- or MonomerPattern to a
    complete ReactionPattern."""

    if isinstance(v, ReactionPattern):
        return v
    else:
        try:
            return ReactionPattern([as_complex_pattern(v)])
        except InvalidComplexPatternException:
            raise InvalidReactionPatternException



class Parameter(Component):

    def __init__(self, name, value=0.0, _export=True):
        Component.__init__(self, name, _export)
        self.value = value

    def __repr__(self):
        return  '%s(name=%s, value=%s)' % (self.__class__.__name__, repr(self.name), repr(self.value))



class Compartment(Component):
    """The Compartment class expects a "name", "parent", "dimension", and "size" variable from the
    compartment call. name is a string, "parent" should be the name of a defined parent, or None. 
    Dimension should be only 2 (e.g. membranes) or 3 (e.g. cytosol). The size units will depend in the
    manner in which the model variable units have been determined. Note, parent is the compartment object.
    example: Compartment('eCell', dimension=3, size=extraSize, parent=None)
    """

    def __init__(self, name, parent=None, dimension=3, size=None, _export=True):
        Component.__init__(self, name, _export)

        if parent != None and isinstance(parent, Compartment) == False:
            raise Exception("parent must be a predefined Compartment or None")
        #FIXME: check for only ONE "None" parent? i.e. only one compartment can have a parent None?

        if size is not None and not isinstance(size, Parameter):
            raise Exception("size must be a parameter (or omitted)")

        self.parent = parent
        self.dimension = dimension
        self.size = size

    def __repr__(self):
        return  '%s(name=%s, parent=%s, dimension=%s, size=%s)' % \
            (self.__class__.__name__, repr(self.name), repr(self.parent), repr(self.dimension), repr(self.size))



class Rule(Component):

    def __init__(self, name, reaction_pattern_set, rate_forward, rate_reverse=None,
                 delete_molecules=False,
                 _export=True):
        Component.__init__(self, name, _export)

        # FIXME: This tuple thing is ugly (used to support >> and <> operators between ReactionPatterns).
        # This is how the reactant and product ReactionPatterns are passed, along with is_reversible.
        if not isinstance(reaction_pattern_set, tuple) and len(reaction_pattern_set) != 3:
            raise Exception("reaction_pattern_set must be a tuple of (ReactionPattern, ReactionPattern, Boolean)")

        try:
            reactant_pattern = as_reaction_pattern(reaction_pattern_set[0])
        except InvalidReactionPatternException as e:
            raise type(e)("Reactant does not look like a reaction pattern")

        try:
            product_pattern = as_reaction_pattern(reaction_pattern_set[1])
        except InvalidReactionPatternException as e:
            raise type(e)("Product does not look like a reaction pattern")

        self.is_reversible = reaction_pattern_set[2]

        if not isinstance(rate_forward, Parameter):
            raise Exception("Forward rate must be a Parameter")
        if self.is_reversible and not isinstance(rate_reverse, Parameter):
            raise Exception("Reverse rate must be a Parameter")

        self.reactant_pattern = reactant_pattern
        self.product_pattern = product_pattern
        self.rate_forward = rate_forward
        self.rate_reverse = rate_reverse
        self.delete_molecules = delete_molecules
        # TODO: ensure all numbered sites are referenced exactly twice within each of reactants and products

    def __repr__(self):
        ret = '%s(name=%s, reactants=%s, products=%s, rate_forward=%s' % \
            (self.__class__.__name__, repr(self.name), repr(self.reactant_pattern), repr(self.product_pattern), repr(self.rate_forward))
        if self.is_reversible:
            ret += ', rate_reverse=%s' % repr(self.rate_reverse)
        if self.delete_molecules:
            ret += ', delete_molecules=True'
        ret += ')'
        return ret



class InvalidComplexPatternException(Exception):
    pass

class InvalidReactionPatternException(Exception):
    pass

class ModelExistsWarning(UserWarning):
    """Issued by Model constructor when a second model is defined."""
    pass

class SymbolExistsWarning(UserWarning):
    """Issued by model component constructors when a name is reused."""
    pass

class InvalidComponentNameError(ValueError):
    """Issued by Component.__init__ when the given name is not valid."""
    def __init__(self, name):
        ValueError.__init__(self, "Not a valid component name: '%s'" % name)



class ComponentSet(collections.Set, collections.Mapping, collections.Sequence):
    """An add-and-read-only container for storing model Components. It behaves mostly like an
    ordered set, but components can also be retrieved by name *or* index by using the [] operator
    (like a dict or list). Components may not be removed or replaced."""
    # The implementation is based on a list instead of a linked list (as OrderedSet is), since we
    # only allow add and retrieve, not delete.

    def __init__(self, iterable=[]):
        self._elements = []
        self._map = {}
        for value in iterable:
            self.add(value)

    def __iter__(self):
        return iter(self._elements)

    def __contains__(self, c):
        if not isinstance(c, Component):
            raise TypeError("Can only work with Components, got a %s" % type(c))
        return c.name in self._map and self[c.name] is c

    def __len__(self):
        return len(self._elements)

    def add(self, c):
        if c not in self:
            if c.name in self._map:
                raise ComponentDuplicateNameError("Tried to add a component with a duplicate name: %s" % c.name)
            self._elements.append(c)
            self._map[c.name] = c

    def __getitem__(self, key):
        # Must support both Sequence and Mapping behavior. This means stringified integer Mapping
        # keys (like "0") are forbidden, but since all Component names must be valid Python
        # identifiers, integers are ruled out anyway.
        if isinstance(key, int) or isinstance(key, long):
            return self._elements[key]
        else:
            return self._map[key]

    def get(self, key, default=None):
        if isinstance(key, (int, long)):
            raise ValueError("Get is undefined for integer arguments, use [] instead")
        try:
            return self[key]
        except KeyError:
            return default

    def iterkeys(self):
        for c in self:
            yield c.name

    def itervalues(self):
        return self.__iter__()

    def iteritems(self):
        for c in self:
            yield (c.name, c)

    def keys(self):
        return [c.name for c in self]

    def values(self):
        return [c for c in self]

    def items(self):
        return zip(self.keys(), self)

    # We reimplement this because collections.Set's __and__ mixin iterates over other, not
    # self. That implementation ends up retaining the ordering of other, but we'd like to keep the
    # ordering of self instead. We require other to be a ComponentSet too so we know it will support
    # "in" efficiently.
    def __and__(self, other):
        if not isinstance(other, ComponentSet):
            return collections.Set.__and__(self, other)
        return ComponentSet(value for value in self if value in other)

    def __rand__(self, other):
        return self.__and__(other)

    def __ror__(self, other):
        return self.__or__(other)

    def __rxor__(self, other):
        return self.__xor__(other)

    def __repr__(self):
        return '{' + \
            ',\n '.join("'%s': %s" % t for t in self.iteritems()) + \
            '}'


class ComponentDuplicateNameError(ValueError):
    """Issued by ComponentSet.add when a component is added with the
    same name as an existing one."""
    pass


def extract_site_conditions(*args, **kwargs):
    """Handle parsing of MonomerPattern site conditions.
    """
    # enforce site conditions as kwargs or a dict but not both
    if (args and kwargs) or len(args) > 1:
        raise Exception("Site conditions may be specified as EITHER keyword arguments OR a single dict")
    # handle normal cases
    elif args:
        site_conditions = args[0].copy()
    else:
        site_conditions = kwargs
    return site_conditions



ANY = MonomerAny()
WILD = MonomerWild()

warnings.simplefilter('always', ModelExistsWarning)
warnings.simplefilter('always', SymbolExistsWarning)



import pysb
from pysb.generator.bng import BngGenerator
import os
import subprocess
import random
import re
import sympy
from StringIO import StringIO

# not ideal, but it will work for now during development
pkg_path = None
pkg_paths_to_check = [
    '/usr/local/share/BioNetGen',
    'c:/Program Files/BioNetGen'
    ]
for check_path in pkg_paths_to_check:
    for subdir in ('', 'Perl2'):
        path = os.path.join(check_path, subdir)
        if os.access(os.path.join(path, 'BNG2.pl'), os.F_OK):
            pkg_path = path
            break
if pkg_path is None:
    msg = "Could not find BioNetGen installed in one of the following locations:\n    " + \
        '\n    '.join(pkg_paths_to_check)
    raise Exception(msg)

class GenerateNetworkError(RuntimeError):
    pass

generate_network_code = """
begin actions
generate_network({overwrite=>1});
end actions
"""


def generate_network(model, cleanup=True, append_stdout=False):
    """Run a model through BNG's generate_network function and return
    the content of the resulting .net file as a string"""
    gen = BngGenerator(model)
    bng_filename = '%s_%d_%d_temp.bngl' % (model.name, os.getpid(), random.randint(0, 10000))
    net_filename = bng_filename.replace('.bngl', '.net')
    output = StringIO()
    try:
        bng_file = open(bng_filename, 'w')
        bng_file.write(gen.get_content())
        bng_file.write(generate_network_code)
        bng_file.close()
        p = subprocess.Popen(['perl', pkg_path + 'BNG2.pl', bng_filename],
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (p_out, p_err) = p.communicate()
        if p.returncode:
            raise GenerateNetworkError(p_err.rstrip())
        net_file = open(net_filename, 'r')
        output.write(net_file.read())
        if append_stdout:
            output.write("#\n# BioNetGen execution log follows\n# ==========")
            output.write(re.sub(r'(^|\n)', r'\n# ', p_out))
    finally:
        if cleanup:
            for filename in [bng_filename, net_filename]:
                if os.access(filename, os.F_OK):
                    os.unlink(filename)
    return output.getvalue()


def generate_equations(model):
    # only need to do this once
    # TODO track "dirty" state, i.e. "has model been modified?"
    #   or, use a separate "math model" object to contain ODEs
    if model.odes:
        return
    lines = iter(generate_network(model).split('\n'))
    try:
        while 'begin species' not in lines.next():
            pass
        model.species = []
        while True:
            line = lines.next()
            if 'end species' in line: break
            _parse_species(model, line)

        while 'begin reactions' not in lines.next():
            pass
        model.odes = [sympy.S(0)] * len(model.species)
        reaction_cache = {}
        while True:
            line = lines.next()
            if 'end reactions' in line: break
            (number, reactants, products, rate, rule) = line.strip().split()
            # the -1 is to switch from one-based to zero-based indexing
            reactants = tuple(int(r) - 1 for r in reactants.split(','))
            products = tuple(int(p) - 1 for p in products.split(','))
            rate = rate.rsplit('*')
            (rule_name, is_reverse) = re.match(r'#(\w+)(?:\((reverse)\))?', rule).groups()
            is_reverse = bool(is_reverse)
            r_names = ['s%d' % r for r in reactants]
            combined_rate = sympy.Mul(*[sympy.S(t) for t in r_names + rate]) 
            rule = model.rules[rule_name]
            reaction = {
                'reactants': reactants,
                'products': products,
                'rate': combined_rate,
                'rule': rule_name,
                'reverse': is_reverse,
                }
            model.reactions.append(reaction)
            key = (rule_name, reactants, products)
            key_reverse = (rule_name, products, reactants)
            reaction_bd = reaction_cache.get(key_reverse)
            if reaction_bd is None:
                # make a copy of the reaction dict
                reaction_bd = dict(reaction)
                # default to false until we find a matching reverse reaction
                reaction_bd['reversible'] = False
                reaction_cache[key] = reaction_bd
                model.reactions_bidirectional.append(reaction_bd)
            else:
                reaction_bd['reversible'] = True
                reaction_bd['rate'] -= combined_rate
            for p in products:
                model.odes[p] += combined_rate
            for r in reactants:
                model.odes[r] -= combined_rate
        # fix up reactions whose reverse version we saw first
        for r in model.reactions_bidirectional:
            if r['reverse']:
                r['reactants'], r['products'] = r['products'], r['reactants']
                r['rate'] *= -1
            # now the 'reverse' value is no longer needed
            del r['reverse']

        while 'begin groups' not in lines.next():
            pass
        model.observable_groups = {}
        while True:
            line = lines.next()
            if 'end groups' in line: break
            _parse_group(model, line)

    except StopIteration as e:
        pass


def _parse_species(model, line):
    index, species, value = line.strip().split()
    complex_compartment_name, complex_string = re.match(r'(?:@(\w+)::)?(.*)', species).groups()
    monomer_strings = complex_string.split('.')
    monomer_patterns = []
    for ms in monomer_strings:
        monomer_name, site_strings, monomer_compartment_name = re.match(r'(\w+)\(([^)]*)\)(?:@(\w+))?', ms).groups()
        site_conditions = {}
        if len(site_strings):
            for ss in site_strings.split(','):
                # FIXME this should probably be done with regular expressions
                if '!' in ss and '~' in ss:
                    site_name, condition = ss.split('~')
                    state, bond = condition.split('!')
                    if bond == '?':
                        bond = pysb.WILD
                    elif bond == '!':
                        bond = pysb.ANY
                    else:
                        bond = int(bond)
                    condition = (state, bond)
                elif '!' in ss:
                    site_name, condition = ss.split('!', 1)
                    if '!' in condition:
                        condition = [int(c) for c in condition.split('!')]
                    else:
                        condition = int(condition)
                elif '~' in ss:
                    site_name, condition = ss.split('~')
                else:
                    site_name, condition = ss, None
                site_conditions[site_name] = condition
        monomer = model.monomers[monomer_name]
        monomer_compartment = model.compartments.get(monomer_compartment_name)
        mp = pysb.MonomerPattern(monomer, site_conditions, monomer_compartment)
        monomer_patterns.append(mp)

    complex_compartment = model.compartments.get(complex_compartment_name)
    cp = pysb.ComplexPattern(monomer_patterns, complex_compartment)
    model.species.append(cp)


def _parse_group(model, line):
    # values are number (which we ignore), name, and species list
    values = line.strip().split()
    group = []
    if len(values) == 3:
        # combination is a comma separated list of [factor*]speciesnumber
        for product in values[2].split(','):
            terms = product.split('*')
            # if no factor given (just species), insert a factor of 1
            if len(terms) == 1:
                terms.insert(0, 1)
            factor = int(terms[0])
            species = int(terms[1]) - 1  # -1 to change to 0-based indexing
            group.append((factor, species))  
    model.observable_groups[values[1]] = group



import pysb.bng
import numpy
from scipy.integrate import ode
from scipy.weave import inline
import distutils.errors
import sympy
import re


use_inline = False
# try to inline a C statement to see if inline is functional
try:
    inline('int i;', force=1)
    use_inline = True
except distutils.errors.CompileError as e:
    pass

# try to load our pysundials cvode wrapper for scipy.integrate.ode
try:
    import pysb.pysundials_helpers
except ImportError as e:
    print e
    pass

# some sane default options for a few well-known integrators
default_integrator_options = {
    'vode': {
        'method': 'bdf',
        'with_jacobian': True,
        },
    'cvode': {
        'method': 'bdf',
        'iteration': 'newton',
        },
    }

def odesolve(model, t, integrator='vode', **integrator_options):
    pysb.bng.generate_equations(model)
    
    param_subs = dict([ (p.name, p.value) for p in model.parameters ])
    param_values = numpy.array([param_subs[p.name] for p in model.parameters])
    param_indices = dict( (p.name, i) for i, p in enumerate(model.parameters) )

    code_eqs = '\n'.join(['ydot[%d] = %s;' % (i, sympy.ccode(model.odes[i])) for i in range(len(model.odes))])
    code_eqs = re.sub(r's(\d+)', lambda m: 'y[%s]' % (int(m.group(1))), code_eqs)
    for i, p in enumerate(model.parameters):
        code_eqs = re.sub(r'\b(%s)\b' % p.name, 'p[%d]' % i, code_eqs)

    # If we can't use weave.inline to run the C code, compile it as Python code instead for use with
    # exec. Note: C code with array indexing, basic math operations, and pow() just happens to also
    # be valid Python.  If the equations ever have more complex things in them, this might fail.
    if not use_inline:
        code_eqs_py = compile(code_eqs, '<%s odes>' % model.name, 'exec')

    y0 = numpy.zeros((len(model.odes),))
    for cp, ic_param in model.initial_conditions:
        si = model.get_species_index(cp)
        y0[si] = ic_param.value

    def rhs(t, y, p):
        ydot = numpy.empty_like(y)
        # note that the evaluated code sets ydot as a side effect
        if use_inline:
            inline(code_eqs, ['ydot', 't', 'y', 'p']);
        else:
            exec code_eqs_py in locals()
        return ydot

    nspecies = len(model.species)
    obs_names = [name for name, rp in model.observable_patterns]
    rec_names = ['__s%d' % i for i in range(nspecies)] + obs_names
    yout = numpy.ndarray((len(t), len(rec_names)))

    # build integrator options list from our defaults and any kwargs passed to this function
    options = {}
    try:
        options.update(default_integrator_options[integrator])
    except KeyError as e:
        pass
    options.update(integrator_options)

    # perform the actual integration
    integrator = ode(rhs).set_integrator(integrator, **options)
    integrator.set_initial_value(y0, t[0]).set_f_params(param_values)
    yout[0, :nspecies] = y0
    i = 1
    while integrator.successful() and integrator.t < t[-1]:
        integrator.integrate(t[i])
        yout[i, :nspecies] = integrator.y
        i += 1

    for i, name in enumerate(obs_names):
        obs_group = model.observable_groups[name]
        if obs_group:
            factors, species = zip(*obs_group)
            obs_values = (yout[:, species] * factors).sum(1)
        else:
            obs_values = numpy.zeros(yout.shape[0])
        yout[:, nspecies + i] = obs_values

    dtype = zip(rec_names, (yout.dtype,) * len(rec_names))
    yrec = numpy.recarray((yout.shape[0],), dtype=dtype, buf=yout)
    return yrec



from pysb import *

Model()

Compartment('extracellular', dimension=3)
Compartment('membrane', parent=extracellular, dimension=2)

Monomer('egf', ['R'])
Monomer('egfr', ['L', 'D', 'C'])

Parameter('Kf_egfr_egf', 1.2)
Parameter('Kr_egfr_egf', 1.1)
Rule('egfr_egf',
     egfr(L=None) + egf(R=None) <>
     egfr(L=1)    % egf(R=1),
     Kf_egfr_egf, Kr_egfr_egf)

Observe('free_egf', egf(R=None))
Observe('free_egfr', egfr(L=None))
Observe('bound', egf(R=1) % egfr(L=1))

Parameter('egf_0', 6.0)
Parameter('egfr_0', 10.0)
Initial(egf(R=None) ** extracellular, egf_0)
Initial(egfr(L=None, D=None, C=None) ** membrane, egfr_0)


if __name__ == '__main__':
    from pysb.generator.bng import BngGenerator
    gen = BngGenerator(model)
    print gen.get_content()
    print "begin actions"
    print "  generate_network({overwrite=>1});"
    print "end actions"




from pysb.core import *


try:
    import reinteract         # fails if reinteract not installed
    reinteract.custom_result  # fails if this code is run outside of the reinteract shell
except (ImportError, AttributeError) as e:
    pass                      # silently skip applying the mixin below
else:
    import pysb.reinteract_integration
    pysb.reinteract_integration.apply_mixins()



#!/usr/bin/env python

# FIXME this should use libsbml if available

import pysb
import pysb.bng
import sympy
import re
import sys
import os
from StringIO import StringIO

def run(model):
    output = StringIO()
    pysb.bng.generate_equations(model)

    output.write(
        """<?xml version="1.0" encoding="UTF-8"?>
<sbml xmlns="http://www.sbml.org/sbml/level2/version4" level="2" version="4">
    <model>
        <listOfCompartments>
            <compartment id="default" name="default" spatialDimensions="0"/>
        </listOfCompartments>
""")

    ics = [[s, 0] for s in model.species]  # complexpattern, initial value
    for cp, ic_param in model.initial_conditions:
        ics[model.get_species_index(cp)][1] = ic_param.value
    output.write("        <listOfSpecies>\n")
    for i, (cp, value) in enumerate(ics):
        name = str(cp).replace('% ', '._br_')  # CellDesigner does something weird with % in names
        output.write('            <species id="s%d" name="%s" compartment="default" initialAmount="%e"/>\n' % (i, name, value));
    output.write("        </listOfSpecies>\n")

    output.write("        <listOfParameters>\n")
    for i, param in enumerate(model.parameters_rules()):
        output.write('            <parameter id="%s" name="%s" value="%e"/>\n' % (param.name, param.name, param.value));
    output.write("        </listOfParameters>\n")

    output.write("        <listOfReactions>\n")
    for i, reaction in enumerate(model.reactions_bidirectional):
        reversible = str(reaction['reversible']).lower()
        output.write('            <reaction id="r%d" name="r%d" reversible="%s">\n' % (i, i, reversible));
        output.write('                <listOfReactants>\n');
        for species in reaction['reactants']:
            output.write('                    <speciesReference species="s%d"/>\n' % species)
        output.write('                </listOfReactants>\n');
        output.write('                <listOfProducts>\n');
        for species in reaction['products']:
            output.write('                    <speciesReference species="s%d"/>\n' % species)
        output.write('                </listOfProducts>\n');
        mathml = '<math xmlns="http://www.w3.org/1998/Math/MathML">' \
            + sympy.printing.mathml(reaction['rate']) + '</math>'
        output.write('                <kineticLaw>' + mathml + '</kineticLaw>\n');
        output.write('            </reaction>\n');
    output.write("        </listOfReactions>\n")

    output.write("    </model>\n</sbml>\n")
    return output.getvalue()

if __name__ == '__main__':
    # sanity checks on filename
    if len(sys.argv) <= 1:
        raise Exception("You must specify the filename of a model script")
    model_filename = sys.argv[1]
    if not os.path.exists(model_filename):
        raise Exception("File '%s' doesn't exist" % model_filename)
    if not re.search(r'\.py$', model_filename):
        raise Exception("File '%s' is not a .py file" % model_filename)
    sys.path.insert(0, os.path.dirname(model_filename))
    model_name = re.sub(r'\.py$', '', os.path.basename(model_filename))
    # import it
    try:
        # FIXME if the model has the same name as some other "real" module
        # which we use, there will be trouble
        # (use the imp package and import as some safe name?)
        model_module = __import__(model_name)
    except StandardError as e:
        print "Error in model script:\n"
        raise
    # grab the 'model' variable from the module
    try:
        model = model_module.__dict__['model']
    except KeyError:
        raise Exception("File '%s' isn't a model file" % model_filename)
    print run(model)






# example of a model defined without SelfExporter

from pysb import *

SelfExporter.do_export = False
SelfExporter.default_model = Model('dummy')

model = Model('explicit')

L = Monomer('L', ['r'])
R = Monomer('R', ['l'])
model.add_component(L)
model.add_component(R)

kf = Parameter('kf', 1e-6)
L_binds_R = Rule('L_binds_R', L(r=None) + R(l=None) >> L(r=1) % R(l=1), kf)
model.add_component(kf)
model.add_component(L_binds_R)

L_0 = Parameter('L_0', 1e2)
R_0 = Parameter('R_0', 1e4)
model.add_component(L_0)
model.add_component(R_0)
model.initial(L(r=None), L_0)
model.initial(R(l=None), R_0)



from pysb import ComponentSet

__all__ = ['rules_using_parameter']

def rules_using_parameter(model, parameter):
    """Return a ComponentSet of rules in the model which make use of the given parameter"""
    cset = ComponentSet()
    for rule in model.rules:
        if rule.rate_forward is parameter or rule.rate_reverse is parameter:
            cset.add(rule)
    return cset



from pysb import *

Model()

Monomer('Kinase', ['b'])
Monomer('Substrate', ['b', 'y'], {'y': ['U','P']})


def catalyze(enz, sub, site, state1, state2, kf, kr, kc):
    """2-step catalytic process"""
    r1_name = '%s_bind_%s' % (enz.name, sub.name)
    r2_name = '%s_produce_%s' % (enz.name, sub.name)
    E = enz(b=None)
    S = sub({'b': None, site: state1})
    ES = enz(b=1) % sub({'b': 1, site: state1})
    P = sub({'b': None, site: state2})
    Rule(r1_name, E + S <> ES, kf, kr)
    Rule(r2_name, ES >> E + P, kc)


Parameter('kf', 1)
Parameter('kr', 1)
Parameter('kc', 1)
catalyze(Kinase, Substrate, 'y', 'U', 'P', kf, kr, kc)

Parameter('dummy_0', 1)
Initial(Kinase(b=None), dummy_0)
Initial(Substrate(b=None,y='U'), dummy_0)



from pysb import *

# This model demonstrates some more advanced macros. It implements
# pore assembly via sequential addition of identical subunits, and
# transport of cargo through the pore.


# macro definitions

def pore_species(Subunit, site1, site2, size):
    """
    Generate a single species representing a homomeric pore, composed
    of <size> copies of <Subunit> bound together in a ring, with bonds
    formed between <site1> of one unit and <site2> of the next.
    """

    if size == 0:
        raise ValueError("size must be an integer greater than 0")
    if size == 1:
        Pore = Subunit({site1: None, site2: None})
    elif size == 2:
        Pore = Subunit({site1: 1, site2: None}) % Subunit({site1: None, site2: 1})
    else:
        Pore = ComplexPattern([], None, match_once=True)
        for i in range(1, size+1):
            Pore %= Subunit({site1: i, site2: i%size+1})
    return Pore

def pore_assembly(Subunit, site1, site2, size, rates):
    """
    Generate rules to chain identical MonomerPatterns <Subunit> into
    increasingly larger pores of up to <size> units, using sites
    <site1> and <site2> to bind the units to each other.
    """
    rules = []
    for i in range(2, size + 1):
        M = pore_species(Subunit, site1, site2, 1)
        S1 = pore_species(Subunit, site1, site2, i-1)
        S2 = pore_species(Subunit, site1, site2, i)
        rules.append(Rule('%s_pore_assembly_%d' % (Subunit.monomer.name, i),
                          M + S1 <> S2, *rates[i-2]))
    return rules

def pore_transport(Subunit, ssite1, ssite2, min_size, max_size, CargoSource, CargoDest, tsite, rates):
    """
    Generate rules to transport MonomerPattern <CargoSource> to
    <CargoDest> through any of a series of pores of at least
    <min_size> and at most <max_size> subunits, as defined by
    pore_assembly. Uses site <tsite> on both Subunit and CargoSource
    to bind cargo to ONE Subunit during transport. tsite on all other
    Subunits remains empty.
    """
    assert tsite in CargoSource.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(site, CargoSource.monomer.name)
    assert tsite in CargoDest.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(site, CargoDest.monomer.name)

    for i in range(min_size, max_size+1):
        # require all pore subunit <tsite> sites to be empty for Pore match
        Pore = pore_species(Subunit({tsite: None}), ssite1, ssite2, i)

        r1_name = '%s_pore_%d_transport_%s_cplx' % (CargoSource.monomer.name, i, Subunit.monomer.name)
        r2_name = '%s_pore_%d_transport_%s_dssc' % (CargoSource.monomer.name, i, Subunit.monomer.name)

        rule_rates = rates[i-min_size]
        CPore = Pore.copy()
        tbondnum = i + 1
        CPore.monomer_patterns[0].site_conditions[tsite] = tbondnum
        Complex = CPore % CargoSource({tsite: tbondnum})
        Rule(r1_name, Pore + CargoSource({tsite: None}) <> Complex, *rule_rates[0:2])
        Rule(r2_name, Complex >> Pore + CargoDest({tsite: None}), rule_rates[2])


# ========================================

Model()

Monomer('Bax', ['bh3', 'd2', 't'])
Monomer('Smac', ['loc', 't'], {'loc': ['m','c']})

Parameter('Bax_0', 1)
Parameter('Smac_0', 1)

Initial(Bax(bh3=None, d2=None, t=None), Bax_0)
Initial(Smac(loc='m', t=None), Smac_0)

prefix = 'Bax_pore_assembly_'
assembly_rates = [
    [Parameter(prefix + '2_f', 1), Parameter(prefix + '2_r', 1)],
    [Parameter(prefix + '3_f', 1), Parameter(prefix + '3_r', 1)],
    [Parameter(prefix + '4_f', 1), Parameter(prefix + '4_r', 1)]
    ]

prefix = 'Bax_Smac_transport_'
transport_rates = [
    [Parameter(prefix + '4_f', 1), Parameter(prefix + '4_r', 1), Parameter(prefix + '4_c', 1)]
    ]

# specify t=None so the pore can't fall apart while it's bound to cargo
pore_assembly(Bax(t=None), 'bh3', 'd2', 4, assembly_rates)

pore_transport(Bax(), 'bh3', 'd2', 4, 4, Smac(loc='m'), Smac(loc='c'), 't', transport_rates)



#!/usr/bin/env python

from pysb.generator.bng import BngGenerator
from pysb.bng import generate_network
import re
import sys
import os


def run(model):
    return generate_network(model, append_stdout=True)


if __name__ == '__main__':
    # sanity checks on filename
    if len(sys.argv) <= 1:
        raise Exception("You must specify the filename of a model script")
    model_filename = sys.argv[1]
    if not os.path.exists(model_filename):
        raise Exception("File '%s' doesn't exist" % model_filename)
    if not re.search(r'\.py$', model_filename):
        raise Exception("File '%s' is not a .py file" % model_filename)
    sys.path.insert(0, os.path.dirname(model_filename))
    model_name = re.sub(r'\.py$', '', os.path.basename(model_filename))
    # import it
    try:
        # FIXME if the model has the same name as some other "real" module which we use,
        # there will be trouble (use the imp package and import as some safe name?)
        model_module = __import__(model_name)
    except StandardError as e:
        print "Error in model script:\n"
        raise
    # grab the 'model' variable from the module
    try:
        model = model_module.__dict__['model']
    except KeyError:
        raise Exception("File '%s' isn't a model file" % model_filename)
    print run(model)



#!/usr/bin/env python

import pysb
import pysb.bng
import sympy
import re
import sys
import os
from StringIO import StringIO


def run(model):
    output = StringIO()
    pysb.bng.generate_equations(model)

    obs_names = [name for name, rp in model.observable_patterns]

    ic_values = [0] * len(model.odes)
    for cp, ic_param in model.initial_conditions:
        ic_values[model.get_species_index(cp)] = ic_param.value

    # list of "dynamic variables"
    pw_x = ["m = pwAddX(m, 's%d', %e);" % (i, ic_values[i]) for i in range(len(model.odes))]

    # parameters
    pw_k = ["m = pwAddK(m, '%s', %e);" % (p.name, p.value) for p in model.parameters]

    # equations (one for each dynamic variable)
    # Note that we just generate C code, which for basic math expressions
    # is identical to matlab.  We just have to change 'pow' to 'power'.
    # Ideally there would be a matlab formatter for sympy.
    pw_ode = ["m = pwAddODE(m, 's%d', '%s');" % (i, sympy.ccode(model.odes[i])) for i in range(len(model.odes))]
    pw_ode = [re.sub(r'pow(?=\()', 'power', s) for s in pw_ode]

    # observables
    pw_y = ["m = pwAddY(m, '%s', '%s');" % (' + '.join(['%f * s%s' % g for g in model.observable_groups[name]]), name) for name in obs_names]

    output.write('% PottersWheel model definition file\n')
    output.write('%% save as %s.m\n' % model_name)
    output.write('function m = %s()\n' % model_name)
    output.write('\n')
    output.write('m = pwGetEmptyModel();\n')
    output.write('\n')
    output.write('% meta information\n')
    output.write("m.ID          = '%s';\n" % model_name)
    output.write("m.name        = '%s';\n" % model_name)
    output.write("m.description = '';\n")
    output.write("m.authors     = {''};\n")
    output.write("m.dates       = {''};\n")
    output.write("m.type        = 'PW-1-5';\n")
    output.write('\n')
    output.write('% dynamic variables\n')
    for x in pw_x:
        output.write(x)
        output.write('\n')
    output.write('\n')
    output.write('% dynamic parameters\n')
    for k in pw_k:
        output.write(k)
        output.write('\n')
    output.write('\n')
    output.write('% ODEs\n')
    for ode in pw_ode:
        output.write(ode)
        output.write('\n')
    output.write('\n')
    output.write('% observables\n')
    for y in pw_y:
        output.write(y)
        output.write('\n')
    output.write('\n')
    output.write('%% end of PottersWheel model %s\n' % model_name)
    return output.getvalue()

if __name__ == '__main__':
    # sanity checks on filename
    if len(sys.argv) <= 1:
        raise Exception("You must specify the filename of a model script")
    model_filename = sys.argv[1]
    if not os.path.exists(model_filename):
        raise Exception("File '%s' doesn't exist" % model_filename)
    if not re.search(r'\.py$', model_filename):
        raise Exception("File '%s' is not a .py file" % model_filename)
    sys.path.insert(0, os.path.dirname(model_filename))
    model_name = re.sub(r'\.py$', '', os.path.basename(model_filename))
    # import it
    try:
        # FIXME if the model has the same name as some other "real" module which we use, there will be trouble
        # (use the imp package and import as some safe name?)
        model_module = __import__(model_name)
    except StandardError as e:
        print "Error in model script:\n"
        raise
    # grab the 'model' variable from the module
    try:
        model = model_module.__dict__['model']
    except KeyError:
        raise Exception("File '%s' isn't a model file" % model_filename)
    print run(model)



import pysb
import reinteract
import gtk
import math


corner_radius = 10
site_radius = 5
sites_y_pos = 35
sites_y_spacing = 35


class MonomerWidget(gtk.DrawingArea):

    def __init__(self, monomer):
        gtk.DrawingArea.__init__(self)
        self.connect("expose_event", self.expose)
        self.set_size_request(100, sites_y_pos + sites_y_spacing * len(monomer.sites))
        self.monomer = monomer

    def expose(self, widget, event):
        self.draw()
        return False

    def draw(self):
        rect = self.get_allocation()
        context = self.context = self.window.cairo_create();

        context.set_line_width(1)

        # body
        (x1, x2) = (corner_radius + 0.5, rect.width - corner_radius - 0.5)
        (y1, y2) = (corner_radius + 0.5, rect.height - corner_radius - 0.5)
        angles = [v * math.pi/2.0 for v in range(0,4)]
        context.arc(x1, y1, corner_radius, angles[2], angles[3])
        context.arc(x2, y1, corner_radius, angles[3], angles[0])
        context.arc(x2, y2, corner_radius, angles[0], angles[1])
        context.arc(x1, y2, corner_radius, angles[1], angles[2])
        context.close_path()
        context.set_source_rgb(1, 1, 1)
        context.fill_preserve()
        context.set_source_rgb(0, 0, 0)
        context.stroke()

        # label
        context.set_font_size(14)
        context.move_to(corner_radius, context.font_extents()[0] + 3)
        context.show_text(self.monomer.name)
        context.move_to(0, context.font_extents()[0] + 8.5)
        context.rel_line_to(100, 0)
        context.stroke()

        # sites
        context.translate(corner_radius, sites_y_pos)
        for site in self.monomer.sites:
            self.draw_site(site)
            context.translate(0, sites_y_spacing)

    def draw_site(self, site):
        context = self.context
        context.set_source_rgb(0, 0, 0)
        context.arc(site_radius, 0, site_radius, 0, 2*math.pi)
        context.stroke()
        context.move_to(0, site_radius + context.font_extents()[0])
        context.show_text(site)
        context.new_path()





# mix-in with some pysb classes, but only if reinteract is available and running
mixin_pairs = [
    (pysb.Monomer, MonomerWidget),
    ]
def apply_mixins():
    for (base_class, widget_class) in mixin_pairs:
        if reinteract.custom_result.CustomResult not in base_class.__bases__:
            base_class.__bases__ += (reinteract.custom_result.CustomResult,)
            base_class.create_widget = lambda self: widget_class(self)



# Integrates Robertson's example, as defined in robertson.py, and
# plots the trajectories.

from pylab import *
from pysb.integrate import odesolve

from robertson import model

# solve from t=0 to t=40
t = linspace(0, 40)

yrec = odesolve(model, t, rtol=1e-4, atol=[1e-8, 1e-14, 1e-6])

# build a normal array with the values from the recarray
y = yrec.view().reshape(len(yrec), len(yrec[0]))
# plot trajectories, each normalized to the range 0-1
p = plot(t, y / y.max(0))
figlegend(p, ['y1', 'y2', 'y3'], 'upper right')
show()



#!/usr/bin/env python

import pysb
import pysb.bng
import sympy
import re
import sys
import os
from StringIO import StringIO

def run(model):
  output = StringIO()
  pysb.bng.generate_equations(model)

  output.write("% MATLAB model definition file\n")
  output.write('%% save as %s_odes.m\n' % model_name)
  output.write('function out = %s_odes(t, input, param)' % model_name)
  output.write("\n\n")

  c_code_consts = '\n'.join(['%% %s = param(%d) = %s;' % (p.name, i+1, p.value) for i, p in
      enumerate(model.parameters)])
  c_code_eqs = '\n'.join(['out(%d,1) = %s;' % (i+1, sympy.ccode(model.odes[i])) for i in
      range(len(model.odes))])
  c_code_eqs = re.sub(r's(\d+)', lambda m: 'input(%s)' % (int(m.group(1))+1), c_code_eqs)
  c_code_eqs = re.sub(r'pow\(', 'power(', c_code_eqs)
  #c_code = c_code_consts + '\n\n' + c_code_eqs
  c_code = c_code_eqs

  c_code_species = '\n'.join(['%% input(%d) = %s;' % (i+1, s) for i, s in
      enumerate(model.species)])

  for i, p in enumerate(model.parameters):
    c_code = re.sub(r'\b(%s)\b' % p.name, 'param(%d)' % (i+1), c_code)

  output.write(c_code_consts + "\n\n")
  output.write(c_code_species + "\n\n")
  output.write(c_code + "\n\n")
  output.write("end\n")
  return output.getvalue()

if __name__ == '__main__':
    # sanity checks on filename
    if len(sys.argv) <= 1:
        raise Exception("You must specify the filename of a model script")
    model_filename = sys.argv[1]
    if not os.path.exists(model_filename):
        raise Exception("File '%s' doesn't exist" % model_filename)
    if not re.search(r'\.py$', model_filename):
        raise Exception("File '%s' is not a .py file" % model_filename)
    sys.path.insert(0, os.path.dirname(model_filename))
    model_name = re.sub(r'\.py$', '', os.path.basename(model_filename))
    # import it
    try:
        # FIXME if the model has the same name as some other "real" module
        # which we use, there will be trouble
        # (use the imp package and import as some safe name?)
        model_module = __import__(model_name)
    except StandardError as e:
        print "Error in model script:\n"
        raise
    # grab the 'model' variable from the module
    try:
        model = model_module.__dict__['model']
    except KeyError:
        raise Exception("File '%s' isn't a model file" % model_filename)
    print run(model)






from pysb import *


rnas = []

def Nuc(name, sites=[], site_states={}, **kwargs):
    return Monomer(name, sites + ['r5', 'r3'], site_states, **kwargs)

def Rna(*args, **kwargs):
    rna = Nuc(*args, **kwargs)
    rnas.append(rna)
    return rna

def degrade_rna(rna):
    Rule('%s_deg' % rna.name, rna(r3=None) >> NONE(), rna_deg)


Model()

Parameter('rna_deg', 1e-6)
Parameter('a_deg', 1e-3)
Parameter('mRNA_1_init', 50)
Parameter('mRNA_2_init', 10)

Rna('mRNA_1')
Rna('mRNA_2')
Nuc('A')
Monomer('NONE')

for rna in rnas:
    degrade_rna(rna)
Rule('deadenylate', A(r3=None) >> NONE(), a_deg, delete_molecules=True)

Initial(mRNA_1(r5=None, r3=1) % A(r5=1, r3=2) % A(r5=2, r3=None), mRNA_1_init)
Initial(mRNA_2(r5=None, r3=1) % A(r5=1, r3=2) % A(r5=2, r3=3) % A(r5=3, r3=None), mRNA_2_init)




if __name__ == '__main__':
    from pysb.tools.export_bng import run as run_export
    print run_export(model)



#!/usr/bin/env python

import sys
import os
import re
import pygraphviz
import pysb.bng

def run(model):
    pysb.bng.generate_equations(model)
    graph = pygraphviz.AGraph(rankdir="LR")
    for si, cp in enumerate(model.species):
        sgraph_name = 'cluster_s%d' % si
        sgraph = graph.add_subgraph(sgraph_name, label='s%d' % si,
                                    color="gray75", fontsize="20")
        bonds = {}
        for mi, mp in enumerate(cp.monomer_patterns):
            monomer_node = '%s_%d' % (sgraph_name, mi)
            sgraph.add_node(monomer_node,
                            label=mp.monomer.name,
                            shape="rectangle",
                            fillcolor="lightblue", style="filled",
                            fontsize="12",
                            width=".3", height=".3", margin="0.06,0")
            for site in mp.monomer.sites:
                site_state = None
                cond = mp.site_conditions[site]
                if isinstance(cond, str):
                    site_state = cond
                elif isinstance(cond, tuple):
                    site_state = cond[0]
                site_label = site
                if site_state is not None:
                    site_label += '=%s' % site_state
                site_node = '%s_%s' % (monomer_node, site)
                sgraph.add_node(site_node, label=site_label,
                                fontname="courier", fontsize='10',
                                fillcolor="yellow", style="filled", color="transparent",
                                width=".2", height=".2", margin="0")
                sgraph.add_edge(monomer_node, site_node,
                                style="bold")
            for site, value in mp.site_conditions.items():
                site_bonds = []
                if isinstance(value, int):
                    site_bonds.append(value)
                elif isinstance(value, tuple):
                    site_bonds.append(value[1])
                elif isinstance(value, list):
                    site_bonds += value
                for b in site_bonds:
                    bonds.setdefault(b, []).append('%s_%s' % (monomer_node, site))
        for bi, sites in bonds.items():
            sgraph.add_edge(sites, label=str(bi),
                            style="dotted")
    return graph.string()

if __name__ == '__main__':
    # sanity checks on filename
    if len(sys.argv) <= 1:
        raise Exception("You must specify the filename of a model script")
    model_filename = sys.argv[1]
    if not os.path.exists(model_filename):
        raise Exception("File '%s' doesn't exist" % model_filename)
    if not re.search(r'\.py$', model_filename):
        raise Exception("File '%s' is not a .py file" % model_filename)
    sys.path.insert(0, os.path.dirname(model_filename))
    model_name = re.sub(r'\.py$', '', os.path.basename(model_filename))
    # import it
    try:
        # FIXME if the model has the same name as some other "real" module which we use,
        # there will be trouble (use the imp package and import as some safe name?)
        model_module = __import__(model_name)
    except StandardError as e:
        print "Error in model script:\n"
        raise
    # grab the 'model' variable from the module
    try:
        model = model_module.__dict__['model']
    except KeyError:
        raise Exception("File '%s' isn't a model file" % model_filename)
    print run(model)



import pysb
from pysb.generator.kappa import KappaGenerator
import os
import subprocess
import random
import re
import sympy
import numpy as np
import pygraphviz as pgv

### "PUBLIC" Functions ################################################
# i.e., functions that are most useful to be called externally


# Runs the model using kasim with the specified arguments for time and
# number of points (note that it also generates the influence and flux
# maps, though they are not used here).
#
# Returns the kasim simulation data as a numpy array, which can be
# plotted using the plot command.
def get_kasim_data(model, time=10000, points=1000):
  outs = run_kasim(model, time, points)
  return parse_kasim_outfile(outs['out'])


# Runs Kasim with no simulation events, which generates the influence map,
# and then displays it using GraphViz (assumes that GraphViz is set up and is the
# default program for opening .gv files)
def show_influence_map(model):
  kasim_dict = run_kasim(model, time=0, points=0)
  im_filename = kasim_dict['im']
  open_file(im_filename)


# Runs complx with the appropriate arguments for generating the contact map.
# DOESN'T WORK: WHY???
def show_contact_map(model):
  gen = KappaGenerator(model, dialect='complx')
  #kappa_filename = '%d_%d_temp.ka' % (os.getpid(), random.randint(0, 10000))
  kappa_filename = '%s.ka' % model.name
  jpg_filename = kappa_filename.replace('.ka', '.jpg')
  args = ['--output-high-res-contact-map-jpg', jpg_filename]
  run_complx(gen, kappa_filename, args)
  open_file(jpg_filename)


### "PRIVATE" Functions ###############################################

# Generalized method for passing arguments to the complx executable.
def run_complx(gen, kappa_filename, args):
  try:
      kappa_file = open(kappa_filename, 'w')
      kappa_file.write(gen.get_content())
      kappa_file.close()
      cmd = 'complx ' + ' '.join(args) + ' ' + kappa_filename
      print "Command: " + cmd
      p = subprocess.Popen(['complx'] + args + [kappa_filename],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      #p.communicate()
      p.wait()

      if p.returncode:
          #raise Exception(p.stdout.read())
          raise Exception(p.stderr.read())

      #contact_map = pgv.AGraph(dot_filename)
      #contact_map.layout()
      #contact_map.draw(kappa_filename.replace('.ka', '.png'))
  except Exception as e:
      raise Exception("problem running complx: " + str(e))





# Runs kasim, which
def run_kasim(model, time=10000, points=200):
  gen = KappaGenerator(model)
  #kappa_filename = '%d_%d_temp.ka' % (os.getpid(), random.randint(0, 10000))
  kappa_filename = '%s.ka' % model.name
  im_filename = kappa_filename.replace('.ka', '_im.gv')
  fm_filename = kappa_filename.replace('.ka', '_fm.gv')
  out_filename = kappa_filename.replace('.ka', '.out')

  args = ['-i', kappa_filename, '-t', str(time), '-p', str(points),
          '-o', out_filename, '-im', im_filename, '-flux', fm_filename]

  try:
      kappa_file = open(kappa_filename, 'w')
      kappa_file.write(gen.get_content())
      kappa_file.close()
      p = subprocess.Popen(['KaSim'] + args)
                            #stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
      p.communicate()

      if p.returncode:
          raise Exception(p.stdout.read())

  except Exception as e:
      raise Exception("problem running KaSim: " + str(e))
  #finally:
      #for filename in [kappa_filename, dot_filename]:
      #    if os.access(filename, os.F_OK):
      #        os.unlink(filename)

  output_dict = {'out':out_filename, 'im':im_filename, 'fm':'flux.dot'}
  return output_dict

# end run_kasim


# Parses the outputfile produced by kasim, which has the form
def parse_kasim_outfile(out_filename):
  try:
    out_file = open(out_filename, 'r')

    line = out_file.readline().strip() # Get the first line
    out_file.close()
    line = line[2:]  # strip off opening '# '
    raw_names = re.split(' ', line)
    column_names = []

    # Get rid of the quotes surrounding the observable names
    for raw_name in raw_names:
      mo = re.match("'(.*)'", raw_name)
      if (mo): column_names.append(mo.group(1))
      else: column_names.append(raw_name)

    # Create the dtype argument for the numpy record array
    dt = zip(column_names, ('float',)*len(column_names))

    # Load the output file as a numpy record array, skip the name row
    arr = np.loadtxt(out_filename, dtype=dt, skiprows=1)

  except Exception as e:
    raise Exception("problem parsing KaSim outfile: " + str(e))

  return arr


# Utility function for opening files for display (jpg, gv, dot, etc.)
def open_file(filename):
  try:
      p = subprocess.Popen(['open'] + [filename],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      #p.communicate()
      p.wait()
      if p.returncode:
          raise Exception(p.stderr.read())
  except Exception as e:
      raise Exception("Problem opening file: ", e)


# DEPRECATED! Since Kasim always generates the flux and influence maps when run
def generate_influence_map(model):
  gen = KappaGenerator(model)
  kappa_filename = '%d_%d_temp.ka' % (os.getpid(), random.randint(0, 10000))
  dot_filename = kappa_filename.replace('.ka', '.jpg')
  args = ['--output-influence-map-jpg', jpg_filename]
  run_complx(gen, kappa_filename, args)






#!/usr/bin/env python

from distutils.core import setup
import sys, subprocess, traceback, re

def main():

    try:
        version_file = file('RELEASE-VERSION', 'w+')
        version = get_version()
        version_file.seek(0, 0)
        version_file.write(version)
    except Exception as e:
        try:
            version = version_file.read()
        except Exception as e:
            sys.stderr.write(str(e))
            return

    setup(name='pysb',
          version=version,
          description='Python Systems Biology modeling framework',
          long_description='PySB (pronounced "Pie Ess Bee") is a framework ' + \
              'for building rule-based mathematical models of biochemical ' + \
              'systems. It works nicely with scientific Python libraries ' + \
              'such as NumPy, SciPy and SymPy for model simulation and ' + \
              'analysis.',
          author='Jeremy Muhlich',
          author_email='jmuhlich@bitflood.org',
          url='http://pysb.org/',
          packages=['pysb', 'pysb.generator', 'pysb.tools', 'pysb.examples'],
          keywords=['systems', 'biology', 'model', 'rules'],
          classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: BSD License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 2',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Topic :: Scientific/Engineering :: Chemistry',
            'Topic :: Scientific/Engineering :: Mathematics',
            ],
          )

class GitError(Exception):
    pass

def get_version():
    """Get a nice version number from git-describe"""
    gitcmd = ['git', 'describe', '--always', '--abbrev=4']
    try:
        gitproc = subprocess.Popen(gitcmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        retcode = gitproc.wait()
        if retcode:
            raise GitError(gitproc.stderr.read())
        version = gitproc.stdout.next().strip()
        version = re.sub(r'^v', '', version)
        return version
    except (OSError, GitError) as e:
        raise Exception("Error running 'git describe' to determine version:\n\n" +
                        "command\n=====\n" + " ".join(gitcmd) + "\n\n" +
                        "error\n====\n" + str(e) + "\n")

if __name__ == '__main__':
    main()



from scipy.integrate.ode import IntegratorBase
from pysundials import cvode as _cvode, nvecserial
import ctypes

def cvode_rhs_func(t, y, ydot, f_data):
    # take the values that cvode passes its rhs, and adapt them to what
    # scipy.integrate.ode's rhs function expects
    y = y.asarray()
    f_data = ctypes.py_object.from_address(f_data).value
    (f, f_params) = f_data[:]
    ydot[:] = f(t, y, *f_params)[:]
    return 0

class cvode(IntegratorBase):
    valid_methods = {
        'adams': _cvode.CV_ADAMS,
        'bdf': _cvode.CV_BDF,
        }
    valid_iterations = {
        'functional': _cvode.CV_FUNCTIONAL,
        'newton': _cvode.CV_NEWTON,
        }

    def __init__(self, method='adams', iteration='functional', rtol=1.0e-6, atol=1.0e-12):
        if method not in cvode.valid_methods:
            raise Exception("%s is not a valid value for method -- please use one of the following: %s" %
                            (method, [m for m in cvode.valid_methods]))
        if iteration not in cvode.valid_iterations:
            raise Exception("%s is not a valid value for iteration -- please use one of the following: %s" %
                            (iteration, [m for m in cvode.valid_iterations]))
        self.method = method
        self.iteration = iteration
        self.rtol = rtol
        self.atol = atol
        self.first_step = True

    def reset(self, n, has_jac):
        if has_jac:
            raise Exception("has_jac not yet supported")
        self.success = 1
        self.y = _cvode.NVector([0] * n)
        self.first_step = True
        # initialize the cvode memory object
        self.cvode_mem = _cvode.CVodeCreate(cvode.valid_methods[self.method],
                                            cvode.valid_iterations[self.iteration])
        # allocate memory for cvode (even though we need to call CVodeReAlloc right away on the
        # first call to run(), as it seems to be in the spirit of the scipy.integrate.ode design
        # that memory allocation happens here)
        # note that we don't know t0 here, so we pass 0.0 and reinit it later.
        _cvode.CVodeMalloc(self.cvode_mem, cvode_rhs_func, 0.0, self.y, _cvode.CV_SS, self.rtol, self.atol)
        # initialize the dense linear solver
        _cvode.CVDense(self.cvode_mem, n)

    def run(self, f, jac, y0, t0, t1, f_params, jac_params):
        if self.first_step:
            # copy initial state from input ndarray y0 to our nvector y
            self.y[:] = y0[:]
            # reinitialize cvode, now that we have func and t0, and actual values for y0
            _cvode.CVodeReInit(self.cvode_mem, cvode_rhs_func, t0, self.y, _cvode.CV_SS, self.rtol, self.atol)
            # tell cvode about our rhs function's user data
            f_data = ctypes.cast(ctypes.pointer(ctypes.py_object((f, f_params))), ctypes.c_void_p)
            _cvode.CVodeSetFdata(self.cvode_mem, f_data)
        tret = _cvode.realtype()
        flag = _cvode.CVode(self.cvode_mem, t1, self.y, tret, _cvode.CV_NORMAL)
        if flag < 0:
            self.success = 0
            print "cvodes error: %d (see SUNDIALS manual for more information)" % flag
        return (self.y, t1)

IntegratorBase.integrator_classes.append(cvode)



#!/usr/bin/env python

import pysb
import pysb.bng
import sympy
import re
import sys
import os
from StringIO import StringIO
from math import floor, log

def run(model):
  output = StringIO()
  pysb.bng.generate_equations(model)

  output.write("(* Mathematica model definition file *)\n")
  output.write("(* Model Name: " + model.name + " *)\n\n")
  output.write("(*\n")
  output.write("Run with, for example:\n")
  output.write("tmax = 10\n")
  output.write("soln = NDSolve[Join[eqns, initconds], slist, {t, 0, tmax}]\n")
  output.write("Plot[s0[t] /. soln, {t, 0, tmax}, PlotRange -> All]\n")
  output.write("*)\n\n")

  #output.write('(* save as %s_odes.txt *)\n' % model_name)
  #output.write("\n\n")

  # PARAMETERS
  # Note that in Mathematica, underscores are not allowed in variable names, so
  # we simply strip them out here
  c_code_consts = '' 
  for i, p in enumerate(model.parameters):
    pname = p.name.replace('_', '')
    exp = floor(log(p.value, 10))

    if (not exp == 0):
      c_code_consts += '%s = %g * 10^%g;\n' %  (pname, p.value / 10**exp, exp)
    else:
      c_code_consts += '%s = %g;\n' %  (pname, p.value)

  ## ODEs ###
  c_code_eqs = 'eqns = {\n'
  # Concatenate the equations
  c_code_eqs += ',\n'.join(['s%d == %s' % (i, sympy.ccode(model.odes[i]))
                           for i in range(len(model.odes))])
  # Replace, e.g., s0 with s[0]
  c_code_eqs = re.sub(r's(\d+)', lambda m: 's%s[t]' % (int(m.group(1))),
                      c_code_eqs)
  # Add the derivative symbol ' to the left hand sides
  c_code_eqs = re.sub(r's(\d+)\[t\] ==', r"s\1'[t] ==", c_code_eqs)
  # Correct the exponentiation syntax
  c_code_eqs = re.sub(r'pow\((.*), (.*)\)', r'\1^\2', c_code_eqs)
  c_code_eqs += '\n}'
  #c_code = c_code_eqs
  # Eliminate underscores from parameter names in equations
  for i, p in enumerate(model.parameters):
    c_code_eqs = re.sub(r'\b(%s)\b' % p.name, p.name.replace('_', ''), c_code_eqs)

  ## INITIAL CONDITIONS
  ic_values = ['0'] * len(model.odes)
  for i, ic in enumerate(model.initial_conditions):
    ic_values[model.get_species_index(ic[0])] = ic[1].name.replace('_', '')

  c_code_ics = 'initconds = {\n'
  c_code_ics += ',\n'.join(['s%s[0] == %s' % (i, val) for i, val in enumerate(ic_values)])
  c_code_ics += '\n}'
  
  ## SOLVE LIST
  c_code_slist = 'slist = {\n'
  c_code_slist += ',\n'.join(['s%s[t]' % (i) for i in range(0, len(model.odes))])
  c_code_slist += '\n}'

  ## OBSERVABLES
  c_code_obs = ''
  for obs_name in model.observable_groups:
    c_code_obs += obs_name + ' = '
    groups = model.observable_groups[obs_name]
    c_code_obs += '+'.join(['s%s[t]' % (g[1]) for g in groups]) 
    c_code_obs += ' /. soln\n' 
  
  # Add comments identifying the species
  c_code_species = '\n'.join(['(* s%d[t] = %s *)' % (i, s) for i, s in
                        enumerate(model.species)])

  output.write('(* Parameters *)\n')
  output.write(c_code_consts + "\n\n")
  output.write('(* List of Species *)\n')
  output.write(c_code_species + "\n\n")
  output.write('(* ODEs *)\n')
  output.write(c_code_eqs + "\n\n")
  output.write('(* Initial Conditions *)\n')
  output.write(c_code_ics + "\n\n")
  output.write('(* List of Variables (e.g., as an argument to NDSolve) *)\n')
  output.write(c_code_slist + '\n\n')
  output.write('(* Run the simulation *)\n')
  output.write('tmax = 100\n')
  output.write('soln = NDSolve[Join[eqns, initconds], slist, {t, 0, tmax}]\n\n')
  output.write('(* Observables *)\n')
  output.write(c_code_obs + '\n\n')

  #output.write("end\n")
  return output.getvalue()

if __name__ == '__main__':
    # sanity checks on filename
    if len(sys.argv) <= 1:
        raise Exception("You must specify the filename of a model script")
    model_filename = sys.argv[1]
    if not os.path.exists(model_filename):
        raise Exception("File '%s' doesn't exist" % model_filename)
    if not re.search(r'\.py$', model_filename):
        raise Exception("File '%s' is not a .py file" % model_filename)
    sys.path.insert(0, os.path.dirname(model_filename))
    model_name = re.sub(r'\.py$', '', os.path.basename(model_filename))
    # import it
    try:
        # FIXME if the model has the same name as some other "real" module
        # which we use, there will be trouble
        # (use the imp package and import as some safe name?)
        model_module = __import__(model_name)
    except StandardError as e:
        print "Error in model script:\n"
        raise
    # grab the 'model' variable from the module
    try:
        model = model_module.__dict__['model']
    except KeyError:
        raise Exception("File '%s' isn't a model file" % model_filename)
    print run(model)






from pysb import *

# Demonstration of the proper representation of BAX-style tetramers,
# where the subunits dimerize and then tetramerize instead of
# assembling sequentially.

Model()

# Each BAX-BAX bond must always involve a t1 site on one monomer and a
# t2 site on the other.
Monomer('BAX', ['t1', 't2', 'inh'])
Monomer('MCL1', ['b'])

# Two lone monomers form a dimer.
Parameter('kdimf', 1e-6)
Parameter('kdimr', 1e-7)
Rule('bax_dim',
     BAX(t1=None, t2=None) + BAX(t1=None, t2=None) <>
     BAX(t1=1, t2=None) % BAX(t1=None, t2=1),
     kdimf, kdimr)

# Two lone dimers form a tetramer, with a higher rate than the dimerization.
Parameter('ktetf', 1e-3)
Parameter('ktetr', 1e-4)
Rule('bax_tet',
     BAX(t1=1, t2=None) % BAX(t1=None, t2=1) + BAX(t1=2, t2=None) % BAX(t1=None, t2=2) <>
     BAX(t1=1, t2=3) % BAX(t1=4, t2=1) % BAX(t1=2, t2=4) % BAX(t1=3, t2=2),
     ktetf, ktetr)

# An inhibitory protein can bind to a BAX subunit at any time.
Parameter('kbaxmcl1f', 1e-5)
Parameter('kbaxmcl1r', 1e-6)
Rule('bax_inh_mcl1',
     BAX(inh=None) + MCL1(b=None) <>
     BAX(inh=1)    % MCL1(b=1),
     kbaxmcl1f, kbaxmcl1r)

# Initial conditions
Parameter('BAX_0', 1e3)
Initial(BAX(t1=None, t2=None, inh=None), BAX_0)
Parameter('MCL1_0', 1e2)
Initial(MCL1(b=None), MCL1_0)

# We must fully specify all four BAX-BAX bonds, otherwise the pattern
# is too loose, match a given species multiple times (beyond the
# factor of four expected due to the rotational symmetry of the
# tetramer), resulting in erroneously high values.
Observe('BAX4', BAX(t1=1, t2=3) % BAX(t1=4, t2=1) % BAX(t1=2, t2=4) % BAX(t1=3, t2=2))
# Same all-bonds requirement here.  However since the BAX tetramer is
# considered inhibited when even one subunit has an inhibitor bound,
# we only need to explicitly write inh=ANY on one of the monomer
# patterns.
Observe('BAX4_inh', BAX(inh=ANY, t1=1, t2=3) % BAX(t1=4, t2=1) % BAX(t1=2, t2=4) % BAX(t1=3, t2=2))


if __name__ == '__main__':
    from pysb.generator.bng import BngGenerator
    gen = BngGenerator(model)
    print gen.get_content()
    print ""
    print "begin actions"
    print "  generate_network({overwrite=>1});"
    print "  simulate_ode({t_end=>21600,n_steps=>360});" # 6 hours, 1-minute steps
    print "end actions"



import inspect
from pysb import *

DEFAULT_UNI_KF = 1e6
DEFAULT_BI_KF = 1e6
DEFAULT_KR = 0.1
DEFAULT_KC = 1

def alias_model_components(model=None):
    """Make all model components visible as symbols in the caller's global namespace"""
    if model is None:
        model = SelfExporter.default_model
    caller_globals = inspect.currentframe().f_back.f_globals
    components = dict((c.name, c) for c in model.all_components())
    caller_globals.update(components)

def two_step_mod_sites(enz, enz_site, sub, sub_site, prod, klist=None):
    """Automation of the Enz + Sub <> Enz:Sub >> Enz + Prod two-step catalytic reaction.
    This function assumes that there is a site named 'bf' (bind site for fxn)
    which it uses by default. This also assumes Enz returns to its original state.
    In an aim for simplicity, site 'bf' need not be passed when calling the function by
    the reactants, but THE FULL STATE OF THE PRODUCT must be specified"""
    
    # FIXME: this will fail if the argument passed is a Complex object. 
    r1_name = 'cplx_%s%s_%s%s' % (sub.monomer.name, ''.join(filter(lambda a: a != None, sub.site_conditions.values())),
                                     enz.monomer.name, ''.join(filter(lambda a: a != None, enz.site_conditions.values())))
    r2_name = 'diss_%s%s_via_%s%s' % (prod.monomer.name, ''.join(filter(lambda a: a != None, prod.site_conditions.values())),
                                         enz.monomer.name, ''.join(filter(lambda a: a != None, enz.site_conditions.values())))

    # Default names for the parameters
    if (not klist):
        kf = Parameter(r1_name + '_kf', DEFAULT_BI_KF)
        kr = Parameter(r1_name + '_kr', DEFAULT_KR)
        kc = Parameter(r2_name + '_kc', DEFAULT_KC)
    else:
        kf, kr, kc = klist #forward, reverse, catalytic
     
    assert enz_site in enz.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(enz_site, Enz.monomer.name)
    assert sub_site in sub.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(sub_site, Sub.monomer.name)
    assert prod.is_concrete(), \
        "The product %s is not a fully specified (concrete) pattern." % (prod) 

    # make the intermediate complex components
    etmpdict = enz.site_conditions.copy()
    stmpdict = sub.site_conditions.copy()
    
    etmpdict[enz_site] = 1
    stmpdict[sub_site] = 1

    enz_cplx = enz.monomer(etmpdict)
    sub_cplx = sub.monomer(stmpdict)

    # add the 'bf' site to the patterns
    enz.site_conditions[enz_site] = None
    sub.site_conditions[sub_site] = None

    # now that we have the complex elements formed we can write the first step rule
    Rule(r1_name, enz + sub <> enz_cplx % sub_cplx, kf, kr)
    
    # and finally the rule for the catalytic transformation
    Rule(r2_name, enz_cplx % sub_cplx >> enz + prod, kc)

def two_step_mod(Enz, Sub, Prod, klist=None, site='bf'):
    two_step_mod_sites(Enz, site, Sub, site, Prod, klist)

def two_step_conv(Sub1, Sub2, Prod, klist, site='bf'):
    """Automation of the Sub1 + Sub2 <> Sub1:Sub2 >> Prod two-step reaction (i.e. dimerization).
    This function assumes that there is a site named 'bf' (bind site for fxn)
    which it uses by default. Site 'bf' need not be passed when calling the function."""

    kf, kr, kc = klist
    
    r1_name = 'cplx_%s%s_%s%s' % (Sub2.monomer.name, ''.join(filter(lambda a: a != None, Sub2.site_conditions.values())),
                                     Sub1.monomer.name, ''.join(filter(lambda a: a != None, Sub1.site_conditions.values())))

    #FIXME: this is a bit dirty but it fixes the problem when prod is a pattern
    if isinstance(Prod, MonomerPattern):
        r2_name = 'cplx_%s_via_%s__%s' % (Prod.monomer.name, Sub1.monomer.name, Sub2.monomer.name)
    elif isinstance(Prod, ComplexPattern):
        r2_name = 'cplx_%s_via_%s__%s' % (("_".join([sub.monomer.name for sub in Prod.monomer_patterns])),
                                          Sub1.monomer.name, Sub2.monomer.name)
    
    assert site in Sub1.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(site, Sub1.monomer.name)
    assert site in Sub2.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(site, Sub2.monomer.name)

    # make the intermediate complex components
    s1tmpdict = Sub1.site_conditions.copy()
    s2tmpdict = Sub2.site_conditions.copy()
    
    s1tmpdict[site] = 1
    s2tmpdict[site] = 1

    Sub1Cplx = Sub1.monomer(s1tmpdict)
    Sub2Cplx = Sub2.monomer(s2tmpdict)

    # add the site to the patterns
    Sub1.site_conditions[site] = None
    Sub2.site_conditions[site] = None

    # now that we have the complex elements formed we can write the first step rule
    Rule(r1_name, Sub1 + Sub2 <> Sub1Cplx % Sub2Cplx, kf, kr)
    
    # and finally the rule for the catalytic transformation
    Rule(r2_name, Sub1Cplx % Sub2Cplx >> Prod, kc)

def simple_dim(Sub, Prod, klist, site='bf'):
    """ Convert two Sub species into one Prod species:
    Sub + Sub <> Prod
    """
    kf, kr = klist
    r1_name = 'dimer_%s_to_%s'%(Sub.monomer.name, Prod.monomer.name)
    assert site in Sub.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(site, Sub.monomer.name)

    # create the sites for the monomers
    Sub.site_conditions[site] = None

    # combine the monomers into a product step rule
    Rule(r1_name, Sub + Sub <> Prod, kf, kr)

def pore_species(Subunit, size):
    """
    Generate a single species representing a homomeric pore, composed
    of <size> copies of <Subunit> bound together in a ring, with bonds
    formed between bh3 of one unit and d2 of the next.
    """

    #FIXME: the sites here are hard-coded and named _bh3_ and _d2_
    #not generic and perhaps misleading?
    M = Subunit.monomer
    sc = Subunit.site_conditions
    if size == 0:
        raise ValueError("size must be an integer greater than 0")
    if size == 1:
        Pore = M(sc, bh3=None, d2=None)
    elif size == 2:
        Pore = M(sc, bh3=1, d2=None) % M(sc, d2=1, bh3=None)
    else:
        Pore = ComplexPattern([], None, match_once=True)
        for i in range(1, size+1):
            Pore %= M(sc, bh3=i, d2=i%size+1)
    return Pore

def pore_assembly(Subunit, size, rates):
    """
    Generate rules to chain identical MonomerPatterns <Subunit> into
    increasingly larger pores of up to <size> units, using sites bh3
    and d2 to bind the units to each other.
    """
    rules = []
    for i in range(2, size + 1):
        M = pore_species(Subunit, 1)
        S1 = pore_species(Subunit, i-1)
        S2 = pore_species(Subunit, i)
        rules.append(Rule('%s_pore_assembly_%d' % (Subunit.monomer.name, i),
                          M + S1 <> S2, *rates[i-2]))
    return rules

def pore_transport(Subunit, Source, Dest, min_size, max_size, rates, site='bf'):
    """
    Generate rules to transport MonomerPattern <Source> to <Dest>
    through any of a series of pores of at least <min_size> and at
    most <max_size> subunits, as defined by pore_assembly.  Implicitly
    uses site 'bf' on both Subunit and Source to bind to each other.
    """
    assert site in Source.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(site, Source.monomer.name)
    assert site in Dest.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(site, Dest.monomer.name)

    for i in range(min_size, max_size+1):
        Pore = pore_species(Subunit, i)
        # require all pore subunit bf sites to be empty for Pore match
        for mp in Pore.monomer_patterns:
            mp.site_conditions[site] = None
        SM = Source.monomer
        ssc = Source.site_conditions
        DM = Dest.monomer
        dsc = Dest.site_conditions

        r1_name = '%s_pore_%d_transport_%s_cplx' % (SM.name, i, Subunit.monomer.name)
        r2_name = '%s_pore_%d_transport_%s_dssc' % (SM.name, i, Subunit.monomer.name)

        rule_rates = rates[i-min_size]
        CPore = Pore._copy()
        source_bonds = range(i+1, i+1+i)
        for b in range(i):
            CPore.monomer_patterns[b].site_conditions[site] = source_bonds[b]
        Complex = CPore % SM(ssc, bf=source_bonds)
        Rule(r1_name, Pore + SM(ssc, bf=None) <> Complex, *rule_rates[0:2])
        Rule(r2_name, Complex >> Pore + DM(dsc, bf=None), rule_rates[2])

def one_step_conv(Sub1, Sub2, Prod, klist, site='bf'):
    """ Convert two Sub species into one Prod species:
    Sub + Sub <> Prod
    """
    kf, kr = klist
    r1_name = 'conv_%s_%s_to_%s'%(Sub1.monomer.name, Sub2.monomer.name, Prod.monomer.name)
    assert site in Sub1.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(site, Sub.monomer.name)
    assert site in Sub2.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(site, Sub.monomer.name)
    # create the sites for the monomers

    Sub1.site_conditions[site] = None
    Sub2.site_conditions[site] = None

    # combine the monomers into a product step rule
    Rule(r1_name, Sub1 + Sub2 <> Prod, kf, kr)

def bind(Sub1, site1, Sub2, site2, klist=None):
    """Automation of the Sub1 + Sub2 <> Sub1:Sub2 one-step complex formation, but allows
    the binding sites of both species to be specified.
    Note that it expects Sub1 and Sub2 to be MonomerPatterns (not Monomers), and site1
    and site2 strings indicating the names of the binding sites.
    klist is list of Parameter objects. If klist is not provided as an argument,
    the bind function will generate default forward and reverse parameters.
    """

    # FIXME: this will fail if the argument passed is a complex, or a Monomer object... 
    r_name = 'cplx_%s_%s' % (Sub1.monomer.name, Sub2.monomer.name)
    
    assert site1 in Sub1.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(site1, Sub1.monomer.name)
    assert site2 in Sub2.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(site2, Sub2.monomer.name)

    if (not klist):
        # Default parameter values
        # Diffusion limited on rate of 1e6 and offrate of 1e-1 implies 100nM binding
        # FIXME get rid of these magic numbers!!!! Put in a global setting of some kind
        kf = Parameter(r_name + '_kf', DEFAULT_BI_KF)
        kr = Parameter(r_name + '_kr', DEFAULT_KR)
    else: 
        kf, kr = klist

    # create the site conditions for the complex
    s1tmpdict = Sub1.site_conditions.copy()
    s2tmpdict = Sub2.site_conditions.copy()
    
    s1tmpdict[site1] = 1
    s2tmpdict[site2] = 1

    Sub1Cplx = Sub1.monomer(s1tmpdict)
    Sub2Cplx = Sub2.monomer(s2tmpdict)

    # create the sites for the monomers
    Sub1.site_conditions[site1] = None
    Sub2.site_conditions[site2] = None
    # now that we have the complex elements formed we can write the first step rule
    Rule(r_name, Sub1 + Sub2 <> Sub1Cplx % Sub2Cplx, kf, kr)
   
 
def simple_bind(Sub1, Sub2, klist, site='bf'):
    """Automation of the Sub1 + Sub2 <> Sub1:Sub2 one-step complex formation. 
    This function assumes that there is a site named 'bf' which, for simplicity
    need not be passed. Invokes the bind function using the same name for
    each site."""
        
    bind(Sub1, site, Sub2, site, klist)

inhibit = simple_bind #alias for simplebind

#FIXME: pass klist of sorts?
def simple_bind_table(bindtable, parmlist, lmodel, site='bf'):
    """This assumes that the monomers passed are in their desired state without
    the 'bf' site, which will be used for binding.
    bindtable is a list of lists denoting the reactions between two types of reactants
    as follows:

    bindtable[0]: [                     reactypeA0,       reactypeA1...          reactypeAN]
    bindtable[1]: [                           args,             args...               args)]
    bindtable[2]: [reactypeB0, args, 'parmfamA0B0',    'parmfamA1B0'...    'parmfamANB0'   ]
    bindtable[3]: [reactypeB1, args, 'parmfamA0B1',    'parmfamA1B1'...    'parmfamANB1'   ]

    the variable 'lmodel' is the model passed for local lookup of parameter variables
    """

    # parse the list, extract reactants, products and parameter families
    #first line is one set of reactants
    react0 = bindtable[0]
    react0st = bindtable[1]
    react1 = [row[0] for row in bindtable[2:]]
    react1st = [row[1] for row in bindtable[2:]]

    # Notice this makes intrxns of size/index intrxns[react1][react0]
    intrxns = [row[2:] for row in bindtable[2:]]
    
    # Add the bf sites to the reactant states dict
    # NOTE: this will reset the value if it is already set.
    # Build the prod states dict from react dicts, change bf to 1
    prod0st = []
    prod1st = []
    for d in react0st:
        d[site] = None
        prod0st.append(d.copy())
    for d in react1st:
        d[site] = None
        prod1st.append(d.copy())
    for d in prod0st:
        d[site] = 1
    for d in prod1st:
        d[site] = 1
    
    # loop over interactions
    pc = 0 # parameter counter, cheap way of keeping track of which param set in the list to use
    for i in range(0, len(react1)):
        for j in range(0, len(react0)):
            if intrxns[i][j] is True:
                # get the parameters from the parmlist
                kf = parmlist[pc][0]
                kr = parmlist[pc][1]
                # rule name
                rname = 'cplx_%s_%s' % (react1[i].name, react0[j].name)
                # create the rule
                #print "Generating  %s:%s complex"%(react1[i].name, react0[j].name)
                Rule(rname, react1[i](react1st[i]) + react0[j](react0st[j]) <>
                     react1[i](prod1st[i]) % react0[j](prod0st[j]), 
                     kf, kr)
                pc += 1
    if pc != len(parmlist):
        print "WARNING, unassigned parameters from list", parmlist
        print "Assigned",pc,"parameter pairs from a total of", len(parmlist)

def multisite_bind_table(bindtable):
    """This assumes that the monomers passed are in their desired state without
    the sites which will be used for binding.
    bindtable is a list of lists denoting the reactions between two types of reactants
    as follows:

    bindtable[0]: [                     reactypeA0,       reactypeA1...          reactypeAN]
    bindtable[1]: [                           args,             args...               args)]
    bindtable[2]: [reactypeB0, args, (Bs, As, fwdrate, revrate)',             ...                    ]
    bindtable[3]: [reactypeB1, args,                 ,              ...                    ]

    To indicate that no interaction occurs, simply enter None in the bind table
    the variable 'lmodel' is the model passed for local lookup of parameter variables
    """

    # parse the list, extract reactants, products and parameter families
    #first line is one set of reactants
    react_rows = [row[0] for row in bindtable[2:]]
    react_row_states = [row[1] for row in bindtable[2:]]
    react_cols = bindtable[0]
    react_col_states = bindtable[1]

    # Notice this makes intrxns of size/index intrxns[react1][react0]
    intrxns = [row[2:] for row in bindtable[2:]]

    # loop over interactions
    pc = 1 # parameter counter
    rc = 1 # rule counter, easy way of making sure names don't clash #FIXME
    for i in range(0, len(react_rows)):
        for j in range(0, len(react_cols)):
            if intrxns[i][j] is not None:

                # Add the bf sites to the reactant states dict
                # NOTE: this will reset the value if it is already set.
                # Build the prod states dict from react dicts, change bf to 1
                (react_row_site, react_col_site, kf_val, kr_val) = intrxns[i][j]

                # get the parameters from the parmlist
                #kf = parmlist[pc][0]
                #kr = parmlist[pc][1]
                react_row_state = react_row_states[i]
                react_col_state = react_col_states[j]
                prod_col_state = []
                prod_row_state = []
                # The binding sites of reactants should be unbound
                react_row_state[react_row_site] = None
                react_col_state[react_col_site] = None
                # The state of the products should be unchanged except at the site
                prod_row_state = react_row_state.copy()
                prod_col_state = react_col_state.copy()
                prod_row_state[react_row_site] = 1
                prod_col_state[react_col_site] = 1

                # Create the parameters
                kf = Parameter('kf' + str(pc), kf_val)
                kr = Parameter('kr' + str(pc), kr_val)

                # Rule name
                rname = 'cplx_%s_%s_%d' % (react_rows[i].name, react_cols[j].name, rc)
                # Create the rule
                #print "Generating  %s:%s complex"%(react1[i].name, react0[j].name)
                Rule(rname, react_rows[i](react_row_state) + react_cols[j](react_col_state) <>
                            react_rows[i](prod_row_state) % react_cols[j](prod_col_state),
                    kf, kr)
                pc += 1
                rc += 1


def reversibly_translocate(sub, loc1, loc2, klist=None, locname='loc'):
    """Create fwd and reverse rules defining a reversible state transition for
    the monomer given by the MonomerPattern sub from one localization state
    to another.

    Creates two rules with names following the pattern: 'monomer_loc1_to_loc2' and vice versa.

    The function generates a rule with the name following the pattern 
    --  sub is a MonomerPattern specifying the species that translocates. The localization
        state should not be specified here.
    --  loc1 and loc2 are strings specifying the names of the locations
        (e.g., 'c' for cytoplasmic, 'm' for mitochondrial)
    --  klist is a list of Parameter objects specifying the forward (loc1 to loc2)
        and reverse (loc2 to loc1) rates. If not specified, the parameters are generated
        according to the pattern 'sub_loc1_to_loc2_rate' and 'sub_loc2_to_loc1_rate'.
    --  locname is an optional string specifying the name of the site that describes
        the location. Defaults to 'loc'.
    """

    # FIXME: this will fail if the argument passed is a complex, or a Monomer object... 
    r_name_fwd = '%s_%s_to_%s' % (sub.monomer.name, loc1, loc2)
    r_name_rev = '%s_%s_to_%s' % (sub.monomer.name, loc2, loc1)
    
    # FIXME: ideally, we should also make sure that the localizations themselves
    # have been declared, not just the name of the loc site
    assert locname in sub.monomer.sites_dict, \
        "Required site %s not present in %s as required"%(locname, sub.monomer.name)

    if (not klist):
        # Default parameter values
        # Diffusion limited on rate of 1e6 and offrate of 1e-1 implies 100nM binding
        # FIXME get rid of these magic numbers!!!! Put in a global setting of some kind
        kf = Parameter(r_name_fwd + '_rate', DEFAULT_UNI_KF)
        kr = Parameter(r_name_rev + '_rate', DEFAULT_KR)
    else: 
        kf, kr = klist

    # create the site conditions for the complex
    s1tmpdict = sub.site_conditions.copy()
    s2tmpdict = sub.site_conditions.copy()
 
    sub_loc1 = sub.monomer(s1tmpdict)
    sub_loc2 = sub.monomer(s2tmpdict)

    # specify the localizations for the monomers
    sub_loc1.site_conditions[locname] = loc1
    sub_loc2.site_conditions[locname] = loc2

    # now that we have the complex elements formed we can write the first step rule
    Rule(r_name_fwd, sub_loc1 >> sub_loc2, kf)
    Rule(r_name_rev, sub_loc2 >> sub_loc1, kr)



from pysb import *

Model()

Monomer('Kinase', ['b'])
Monomer('Inhibitor', ['b'])


def inhibit(targ, inh, kf, kr):
    """inhibition by complexation/sequestration"""
    r_name = '%s_inhibited_by_%s' % (targ.name, inh.name)
    T = targ(b=None)
    I = inh(b=None)
    TI = targ(b=1) % inh(b=1)
    Rule(r_name, T + I <> TI, kf, kr)


Parameter('kf', 1)
Parameter('kr', 1)
inhibit(Kinase, Inhibitor, kf, kr)

Parameter('dummy_0', 1)
Initial(Kinase(b=None), dummy_0)
Initial(Inhibitor(b=None), dummy_0)
